# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Member objects (Users with guilds).

.. currentmodule:: curious.dataclasses.member
"""
import collections
import datetime
from typing import List

from curious.dataclasses import guild as dt_guild, role as dt_role, user as dt_user, \
    voice_state as dt_vs
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.permissions import Permissions
from curious.dataclasses.presence import Game, Presence, Status
from curious.exc import HierarchyError, PermissionsError
from curious.util import subclass_builtin, to_datetime


@subclass_builtin(str)
class Nickname(str):
    """
    Represents the nickname of a :class:`.Member`.

    :cvar NONE: A singleton :class:`.Nickname` representing an empty nickname. Equal to the empty \
        string.
    """
    NONE: 'Nickname'

    def __new__(cls, value: str):
        if value is None or value == "":
            try:
                return cls.NONE
            except AttributeError:
                cls.NONE = super().__new__(cls, "")
                return cls.NONE

        return super().__new__(cls, value)

    def __eq__(self, other):
        if other is None and self == self.NONE:
            return True

        return super().__eq__(other)

    def __repr__(self) -> str:
        return f"<Nickname value={super().__repr__()}>"

    async def set(self, new_nickname: str) -> 'Nickname':
        """
        Sets the nickname of the username.

        :param new_nickname: The new nickname of this user. If None, will reset the nickname.
        """
        parent: Member = self.__dict__['parent']

        # Ensure we don't try and set a bad nickname, which makes an empty listener.
        if new_nickname == self:
            return self

        guild: dt_guild.Guild = parent.guild

        me = False
        if parent == parent.guild.me:
            me = True
            if not guild.me.guild_permissions.change_nickname:
                raise PermissionsError("change_nickname")
        else:
            if not guild.me.guild_permissions.manage_nicknames:
                raise PermissionsError("manage_nicknames")

        if parent.top_role >= guild.me.top_role and parent != guild.me:
            raise HierarchyError("Top role is equal to or lower than victim's top role")

        if new_nickname is not None and len(new_nickname) > 32:
            raise ValueError("Nicknames cannot be longer than 32 characters")

        async def _listener(before, after):
            return after.guild == guild and after.id == parent.id

        async with parent._bot.events.wait_for_manager("guild_member_update", _listener):
            await parent._bot.http.change_nickname(guild.id, new_nickname,
                                                   member_id=parent.id, me=me)

        # the wait_for means at this point the nickname has been changed
        return parent.nickname

    async def reset(self) -> 'Nickname':
        """
        Resets a member's nickname.
        """
        return await self.set(self.NONE)


class MemberRoleContainer(collections.Sequence):
    """
    Represents the roles of a :class:`.Member`.
    """

    def __init__(self, member: 'Member'):
        self._member = member

    def _sorted_roles(self) -> 'List[dt_role.Role]':
        if not self._member.guild:
            return []

        roles = filter(
            lambda r: r is not None,
            map(self._member.guild.roles.get, self._member.role_ids)
        )

        return sorted(roles)

    # opt: the default Sequence makes us re-create the sorted role list constantly
    # we don't wanna cache it, without introducing invalidation hell
    # so we just put `__iter__` on `_sorted_roles`
    def __iter__(self) -> type(iter([])):
        return iter(self._sorted_roles())

    def __len__(self) -> int:
        return len(self._member.role_ids)

    def __getitem__(self, item: int):
        return self._sorted_roles()[item]

    @property
    def top_role(self) -> 'dt_role.Role':
        """
        :return: The top :class:`.Role` for this member.
        """
        roles = self._sorted_roles()
        if len(roles) <= 0:
            return self._member.guild.default_role

        return self[0]

    async def add(self, *roles: 'dt_role.Role'):
        """
        Adds roles to this member.

        :param roles: The :class:`.Role` objects to add to this member's role list.
        """

        if not self._member.guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        # Ensure we can add all of these roles.
        for _r in roles:
            if _r >= self._member.guild.me.top_role:
                msg = "Cannot add role {} - it has a higher or equal position to our top role" \
                    .format(_r.name)
                raise HierarchyError(msg)

        async def _listener(before, after: Member):
            if after.id != self._member.id:
                return False

            if not all(role in after.roles for role in roles):
                return False

            return True

        async with self._member._bot.events.wait_for_manager("guild_member_update", _listener):
            role_ids = set([_r.id for _r in self._member.roles] + [_r.id for _r in roles])
            await self._member._bot.http.edit_member_roles(
                self._member.guild_id, self._member.id, role_ids
            )

    async def remove(self, *roles: 'dt_role.Role'):
        """
        Removes roles from this member.

        :param roles: The roles to remove.
        """
        if not self._member.guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        for _r in roles:
            if _r >= self._member.guild.me.top_role:
                msg = "Cannot remove role {} - it has a higher or equal position to our top role" \
                    .format(_r.name)
                raise HierarchyError(msg)

        async def _listener(before, after: Member):
            if after.id != self._member.id:
                return False

            if not all(role not in after.roles for role in roles):
                return False

            return True

        # Calculate the roles to keep.
        to_keep = set(self._member.roles) - set(roles)

        async with self._member._bot.events.wait_for_manager("guild_member_update", _listener):
            role_ids = set([_r.id for _r in to_keep])
            await self._member._bot.http.edit_member_roles(self._member.guild_id, self._member.id,
                                                           role_ids)


class Member(Dataclass):
    """
    A member represents somebody who is inside a guild.
    """

    __slots__ = ("_user_data", "role_ids", "joined_at", "_nickname", "guild_id", "presence",
                 "roles")

    def __init__(self, client, **kwargs):
        super().__init__(kwargs["user"]["id"], client)

        # copy user data for when the user is decached
        self._user_data = kwargs["user"]
        self._bot.state.make_user(self._user_data)

        #: An iterable of role IDs this member has.
        self.role_ids = [int(rid) for rid in kwargs.get("roles", [])]

        #: A :class:`._MemberRoleContainer` that represents the roles of this member.
        self.roles = MemberRoleContainer(self)

        #: The date the user joined the guild.
        self.joined_at = to_datetime(kwargs.get("joined_at", None))  # type: datetime.datetime

        nick = kwargs.get("nick")
        #: The member's current :class:`.Nickname`.
        self._nickname = Nickname(nick)  # type: Nickname

        #: The ID of the guild that this member is in.
        self.guild_id = None  # type: int

        #: The current :class:`.Presence` of this member.
        self.presence = Presence(status=kwargs.get("status", Status.OFFLINE),
                                 game=kwargs.get("game", None))

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` associated with this member.
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def voice(self) -> 'dt_vs.VoiceState':
        """
        :return: The :class:`.VoiceState` associated with this member.
        """
        try:
            return self.guild._voice_states[self.id]
        except (AttributeError, KeyError):
            return None

    @property
    def nickname(self) -> Nickname:
        """
        Represents a member's nickname.

        :getter: A :class:`._Nickname` for this member.
        :setter: Coerces a string nickname into a :class:`._Nickname`. Do not use.
        """
        return self._nickname

    @nickname.setter
    def nickname(self, value: str):
        if not value:
            self._nickname = Nickname.NONE
            return

        self._nickname = Nickname(value)
        self._nickname.__dict__['parent'] = self

    def __hash__(self) -> int:
        return hash(self.guild_id) + hash(self.user.id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Member):
            return NotImplemented

        return other.guild == self.guild and other.user == self.user

    def _copy(self):
        """
        Copies a member object.
        """
        new_object = object.__new__(self.__class__)  # type: Member
        new_object._bot = self._bot

        new_object.id = self.id
        new_object.role_ids = self.role_ids.copy()
        new_object.joined_at = self.joined_at
        new_object.guild_id = self.guild_id
        new_object.presence = self.presence
        new_object.nickname = self.nickname

        return new_object

    def __del__(self):
        try:
            self._bot.state._check_decache_user(self.id)
        except AttributeError:
            # during shutdown
            pass

    @property
    def user(self) -> 'dt_user.User':
        """
        :return: The underlying :class:`.User` for this member.
        """
        try:
            return self._bot.state._users[self.id]
        except KeyError:
            # don't go through make_user as it'll cache it
            return dt_user.User(self._bot, **self._user_data)

    @property
    def name(self) -> str:
        """
        :return: The computed display name of this user.
        """
        return self.nickname if self.nickname != Nickname.NONE else self.user.username

    @property
    def mention(self) -> str:
        """
        :return: A string that mentions this member. 
        """
        if self.nickname:
            return "<@!{}>".format(self.id)

        return self.user.mention

    @property
    def status(self) -> Status:
        """
        :return: The current :class:`.Status` of this member.
        """
        return self.presence.status if self.presence else Status.OFFLINE

    @property
    def game(self) -> Game:
        """
        :return: The current :class:`.Game` this member is playing.
        """
        if not self.presence:
            return None

        if self.presence.status == Status.OFFLINE:
            return None

        return self.presence.game

    @property
    def colour(self) -> int:
        """
        :return: The computed colour of this user.
        """
        roles = reversed(self.roles)

        # NB: you can abuse discord and edit the defualt role's colour
        # so explicitly check that it isn't the default role, and make sure it has a colour
        # in order to get the correct calculated colour
        roles = filter(lambda role: not role.is_default_role and role.colour, roles)
        try:
            return next(roles).colour
        except StopIteration:
            return 0

    @property
    def top_role(self) -> 'dt_role.Role':
        """
        :return: This member's top-most :class:`.Role`.
        """
        return self.roles.top_role

    @property
    def guild_permissions(self) -> Permissions:
        """
        :return: The calculated guild permissions for a member.
        """
        if self == self.guild.owner:
            return Permissions.all()

        bitfield = 0
        # add the default roles
        bitfield |= self.guild.default_role.permissions.bitfield
        for role in self.roles:
            bitfield |= role.permissions.bitfield

        permissions = Permissions(bitfield)
        if permissions.administrator:
            return Permissions.all()

        return permissions

    # Member methods.
    async def send(self, content: str, *args, **kwargs):
        """
        Sends a message to a member in DM.

        This is a shortcut for :meth:`.User.send`.
        """
        return await self.user.send(content, *args, **kwargs)

    async def ban(self, delete_message_days: int = 7):
        """
        Bans this member from the guild.

        :param delete_message_days: The number of days of messages to delete.
        """
        return await self.guild.bans.add(self, delete_message_days=delete_message_days)

    async def kick(self):
        """
        Kicks this member from the guild.
        """
        return await self.guild.kick(self)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Embed objects.

.. currentmodule:: curious.dataclasses.embed
"""

import datetime

from curious.util import attrdict


class Embed(object):  # not an IDObject! Embeds don't have IDs.
    """
    Represents an Embed object on Discord.
    """

    def __init__(self, *,
                 title: str = None,
                 description: str = None,
                 colour: int = None,
                 type_: str = None,
                 url: str = None,
                 timestamp: str = None,
                 **kwargs):

        def make_attrdict(key: str) -> attrdict:
            if key not in kwargs:
                return attrdict()

            return attrdict(**kwargs[key])

        #: The title of this embed.
        self.title = title

        #: The description of this embed.
        self.description = description

        if colour is None:
            # for passing in from discord
            colour = kwargs.get("color")

        #: The colour of this embed.
        self.colour = colour

        #: The type of this embed.
        self.type_ = type_

        #: The URL for this embed.
        self.url = url

        #: The timestamp for this embed.
        self.timestamp = timestamp  # type: datetime.datetime

        #: The fields for this embed.
        self.fields = []

        #: The footer for this embed.
        self.footer = make_attrdict("footer")

        #: The author of this embed.
        self.author = make_attrdict("author")

        #: The image for this embed.
        self.image = make_attrdict("image")

        #: The video for this embed.
        self.video = make_attrdict("video")

        #: The thumbnail for this embed.
        self.thumbnail = make_attrdict("thumbnail")

    def add_field(self, *, name: str, value: str,
                  inline: bool = True) -> 'Embed':
        """
        Adds a field to the embed.

        :param name: The field name.
        :param value: The field value.
        :param inline: Is this field inline?
        :return: The Embed object.
        """
        if isinstance(name, str) and len(name) == 0:
            raise ValueError("Name must not be empty")

        if isinstance(value, str) and len(value) == 0:
            raise ValueError("Value must not be empty")

        self.fields.append(attrdict({"name": str(name), "value": str(value), "inline": inline}))
        return self

    def set_author(self, *, name: str, url: str = None, icon_url: str = None) -> 'Embed':
        """
        Sets the author of this embed.

        :param name: The name of the author.
        :param url: The URL of the author.
        :param icon_url: The URL of the author icon.
        :return: The Embed object.
        """

        self.author = attrdict(
            name=name
        )

        if url:
            self.author.url = url

        if icon_url:
            self.author.icon_url = icon_url

        return self

    def set_footer(self, *, text: str = None, icon_url: str = None) -> 'Embed':
        """
        Sets the footer of this embed.

        :param text: The footer text of this embed.
        :param icon_url: The icon URL for the footer.
        :return: The Embed object.
        """
        self.footer = attrdict()
        if text:
            self.footer.text = text

        if icon_url:
            self.footer.icon_url = icon_url

        return self

    def set_image(self, *, image_url: str) -> 'Embed':
        """
        Sets the image of this embed.

        :param image_url: The image URL of this embed.
        :return: The Embed object.
        """
        self.image = attrdict()

        if not image_url.startswith("http") or image_url.startswith("attachment://"):
            raise ValueError("Image URLs must start with http[s]")

        if image_url:
            self.image.url = image_url

        return self

    def set_thumbnail(self, *, url: str) -> 'Embed':
        """
        Sets the thumbnail image of this embed.
        
        :param url: The image URL of this thumbnail. 
        :return: The Embed object.
        """
        self.thumbnail = attrdict()
        self.thumbnail.url = url

        return self

    def to_dict(self) -> dict:
        """
        Converts this embed into a flattened dict.
        """
        payload = {
            "type": self.type_ if self.type_ else "rich"
        }

        if self.title:
            payload["title"] = self.title

        if self.description:
            payload["description"] = self.description

        if self.url:
            payload["url"] = self.url

        if self.colour:
            payload["color"] = self.colour  # american spelling

        if self.timestamp:
            payload["timestamp"] = self.timestamp.strftime("%Y-%m-%dT%H:%M:%S.%f")

        # attrdicts can be automatically json dumped easily
        # so we just go and shove these right in there
        if self.footer:
            payload["footer"] = self.footer

        if self.thumbnail:
            payload["thumbnail"] = self.thumbnail

        if self.image:
            payload["image"] = self.image

        if self.author:
            payload["author"] = self.author

        payload["fields"] = self.fields

        return payload



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Converter methods.

.. currentmodule:: curious.commands.converters
"""
from typing import Any, List

import typing_inspect

from curious.commands.exc import ConversionFailedError
from curious.dataclasses.channel import Channel
from curious.dataclasses.member import Member
from curious.dataclasses.role import Role


def convert_member(ann, ctx, arg: str) -> Member:
    """
    Converts an argument into a Member.
    """
    member_id = None
    if arg.startswith("<@") and arg.endswith(">"):
        id = arg[2:-1]
        if id[0] == "!":  # strip nicknames
            id = id[1:]

        try:
            member_id = int(id)
        except ValueError:
            raise ConversionFailedError(ctx, arg, Member, "Invalid member ID")
    elif all(i.isdigit() for i in arg):
        member_id = int(arg)

    if member_id is not None:
        member = ctx.guild.members.get(member_id)
    else:
        member = ctx.guild.search_for_member(full_name=arg)

    if member is None:
        raise ConversionFailedError(ctx, arg, Member, "Could not find Member")

    return member


def convert_channel(ann, ctx, arg: str) -> Channel:
    """
    Converts an argument into a Channel.
    """
    channel_id = None
    if arg.startswith("<#") and arg.endswith(">"):
        try:
            channel_id = int(arg[2:-1])
        except ValueError:
            raise ConversionFailedError(ctx, arg, Channel, "Invalid channel ID")
    elif all(i.isdigit() for i in arg):
        channel_id = int(arg)

    if channel_id is not None:
        channel = ctx.guild.channels.get(channel_id)
    else:
        channel = next(filter(lambda c: c.name == arg, ctx.guild.channels.values()), None)

    if channel is None:
        raise ConversionFailedError(ctx, arg, Channel, "Could not find channel")

    return channel


def convert_role(ann, ctx, arg: str) -> Role:
    """
    Converts an argument into a :class:`.Role`.
    """
    role_id = None
    if arg.startswith("<@&") and arg.endswith(">"):
        try:
            role_id = int(arg[3:-1])
        except ValueError:
            raise ConversionFailedError(ctx, arg, Role, "Invalid role ID")
    elif all(i.isdigit() for i in arg):
        role_id = int(arg)

    if role_id is not None:
        role = ctx.guild.roles.get(role_id)
    else:
        role = next(filter(lambda c: c.name == arg, ctx.guild.roles.values()), None)

    if role is None:
        raise ConversionFailedError(ctx, arg, Role, "Could not find role")

    return role


def convert_int(ann, ctx, arg: str) -> int:
    """
    Converts an argument into an integer.
    """
    try:
        return int(arg, 0)
    except ValueError as e:
        raise ConversionFailedError(ctx, arg, int, "Invalid integer") from e


def convert_float(ann, ctx, arg: str) -> float:
    """
    Converts an argument into a float.
    """
    try:
        return float(arg)
    except ValueError as e:
        raise ConversionFailedError(ctx, arg, float, "Invalid float") from e


def convert_list(ann, ctx, arg: str) -> List[Any]:
    """
    Converts a :class:`typing.List`.
    """
    internal = typing_inspect.get_args(ann)[0]
    converter = ctx._lookup_converter(internal)
    sp = arg.split(" ")
    results = []

    for arg in sp:
        results.append(converter(internal, ctx, arg))

    return results



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes that wrap objects returned by Discord.

A **dataclass** is a class that contains data and allows the user to edit classes. These classes are
stateful,  as in they store state which allows the user to edit them as appropriate.

.. currentmodule:: curious.dataclasses

.. autosummary::
    :toctree: dataclasses

    appinfo
    bases
    channel
    embed
    emoji
    guild
    invite
    member
    message
    permissions
    presence
    reaction
    role
    search
    user
    voice_state
    webhook
    widget

"""



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for plugin objects.

.. currentmodule:: curious.commands.plugin
"""
import inspect
import logging
from collections import OrderedDict

import multio

from curious.core import client as md_client


class PluginMeta(type):
    def __prepare__(*args, **kwargs):
        return OrderedDict()  # 3.6 compat


class Plugin(metaclass=PluginMeta):
    """
    Represents a plugin (a collection of events and commands under one class).
    """
    def __init__(self, client: 'md_client.Client'):
        #: The client for this plugin.
        self.client = client

        #: The task group for this plugin.
        self.task_group = None

    async def load(self) -> None:
        """
        Called when this plugin is loaded.

        By default, this does nothing. It is meant to be overridden to customize behaviour.
        """
        pass

    async def spawn(self, cofunc, *args):
        """
        Spawns a task using this plugin's task group.
        """
        if self.task_group is None:
            # spawn a new task group function
            async def task_group_magic():
                logger = logging.getLogger(__name__)
                try:
                    async with multio.asynclib.task_manager() as tg:
                        self.task_group = tg
                        await multio.asynclib.spawn(tg, cofunc, *args)
                except multio.asynclib.TaskGroupError as e:
                    errors = multio.asynclib.unwrap_taskgrouperror(e)
                    for error in errors:
                        logger.exception("Plugin task group crashed!", exc_info=error)
                except Exception as e:
                    logger.exception("Plugin task group crashed!", exc_info=e)
                finally:
                    self.task_group = None

            await multio.asynclib.spawn(self.client.task_manager, task_group_magic)
        else:
            # spawn using the current one
            await multio.asynclib.spawn(self.task_group, cofunc, *args)

    async def unload(self) -> None:
        """
        Called when this plugin is unloaded.

        By default, this does nothing. It is meant to be overridden to customize behaviour.
        """

    def _get_commands(self) -> list:
        """
        Gets the commands for this plugin.
        """
        return [i[1] for i in inspect.getmembers(self, predicate=lambda i: hasattr(i, "is_cmd"))]



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
The main client class.

This contains a definition for :class:`.Client` which is used to interface primarily with Discord.

.. currentmodule:: curious.core.client
"""

import collections
import enum
import functools
import inspect
import logging
import typing
from types import MappingProxyType

import multio

from curious.core import chunker as md_chunker
from curious.core.event import EventContext, EventManager, event as ev_dec, scan_events
from curious.core.gateway import GatewayHandler, open_websocket
from curious.core.httpclient import HTTPClient
from curious.dataclasses import channel as dt_channel, guild as dt_guild, member as dt_member
from curious.dataclasses.appinfo import AppInfo
from curious.dataclasses.invite import Invite
from curious.dataclasses.message import CHANNEL_REGEX, EMOJI_REGEX, MENTION_REGEX
from curious.dataclasses.presence import Game, Status
from curious.dataclasses.user import BotUser, User
from curious.dataclasses.webhook import Webhook
from curious.dataclasses.widget import Widget
from curious.util import base64ify

logger = logging.getLogger("curious.client")


class BotType(enum.IntEnum):
    """
    An enum that signifies what type of bot this bot is.
    
    This will tell the commands handling how to respond, as well as how to log in.
    """
    #: Regular bot. This signifies that the client should log in as a bot account.
    BOT = 1

    #: User bot. This signifies that the client should log in as a user account.
    USERBOT = 2

    # 4 is reserved

    #: No bot responses. This signifies that the client should respond to ONLY USER messages.
    ONLY_USER = 8

    #: No DMs. This signifies the bot only works in guilds.
    NO_DMS = 16

    #: No guilds. This signifies the bot only works in DMs.
    NO_GUILDS = 32

    #: Self bot. This signifies the bot only responds to itself.
    SELF_BOT = 64


class Client(object):
    """
    The main client class. This is used to interact with Discord.

    To start, you can create an instance of the client by passing it the token you want to use:

    .. code-block:: python3

        cl = Client("my.token.string")

    Registering events can be done with the :meth:`.Client.event` decorator, or alternatively
    manual usage of the :class:`.EventHandler` on :attr:`.Client.events`.

    .. code-block:: python3

        @cl.event("ready")
        async def loaded(ctx: EventContext):
            print("Bot logged in.")

    """
    #: A list of events to ignore the READY status.
    IGNORE_READY = [
        "connect",
        "guild_streamed",
        "guild_chunk",
        "guild_available",
        "guild_sync"
    ]

    def __init__(self, token: str, *,
                 state_klass: type = None,
                 bot_type: int = (BotType.BOT | BotType.ONLY_USER)):
        """
        :param token: The current token for this bot.
        :param state_klass: The class to construct the connection state from.
        :param bot_type: A union of :class:`.BotType` that defines the type of this bot.
        """
        #: The mapping of `shard_id -> gateway` objects.
        self._gateways = {}  # type: typing.MutableMapping[int, GatewayHandler]

        #: The number of shards this client has.
        self.shard_count = 0

        #: The token for the bot.
        self._token = token

        if state_klass is None:
            from curious.core.state import State
            state_klass = State

        #: The current connection state for the bot.
        self.state = state_klass(self)

        #: The bot type for this bot.
        self.bot_type = bot_type

        if self.bot_type & BotType.BOT and self.bot_type & BotType.USERBOT:
            raise ValueError("Bot cannot be a bot and a userbot at the same time")

        #: The current :class:`.EventManager` for this bot.
        self.events = EventManager()
        #: The current :class:`.Chunker` for this bot.
        self.chunker = md_chunker.Chunker(self)
        self.chunker.register_events(self.events)

        self._ready_state = {}

        #: The :class:`.HTTPClient` used for this bot.
        self.http = HTTPClient(self._token, bot=bool(self.bot_type & BotType.BOT))

        #: The cached gateway URL.
        self._gw_url = None  # type: str

        #: The application info for this bot. Instance of :class:`.AppInfo`.
        #: This will be None for user bots.
        self.application_info = None  # type: AppInfo

        #: The task manager used for this bot.
        self.task_manager = None

        for (name, event) in scan_events(self):
            self.events.add_event(event)

    @property
    def user(self) -> BotUser:
        """
        :return: The :class:`.User` that this client is logged in as.
        """
        return self.state._user

    @property
    def guilds(self) -> 'typing.Mapping[int, dt_guild.Guild]':
        """
        :return: A mapping of int -> :class:`.Guild` that this client can see.
        """
        return self.state.guilds

    @property
    def invite_url(self) -> str:
        """
        :return: The invite URL for this bot.
        """
        return "https://discordapp.com/oauth2/authorize?client_id={}&scope=bot".format(
            self.application_info.client_id)

    @property
    def events_handled(self) -> collections.Counter:
        """
        A :class:`collections.Counter` of all events that have been handled since the bot's bootup.
        This can be used to track statistics for events.
         
        .. code-block:: python3
        
            @command()
            async def events(self, ctx: Context):
                '''
                Shows the most common events.
                '''
                
                ev = ctx.bot.events_handled.most_common(3)
                await ctx.channel.messages.send(", ".join("{}: {}".format(*x) for x in ev)
        
        """

        c = collections.Counter()
        for gw in self._gateways.values():
            c.update(gw._dispatches_handled)

        return c

    @property
    def gateways(self) -> 'typing.Mapping[int, GatewayHandler]':
        """
        :return: A read-only view of the current gateways for this client. 
        """
        return MappingProxyType(self._gateways)

    def find_channel(self, channel_id: int):
        """
        Finds a channel by channel ID.
        """
        return self.state.find_channel(channel_id)

    async def get_gateway_url(self) -> str:
        """
        :return: The gateway URL for this bot.
        """
        if self._gw_url:
            return self._gw_url

        self._gw_url = await self.http.get_gateway_url()
        return self._gw_url

    async def get_shard_count(self) -> int:
        """
        :return: The shard count recommended for this bot.
        """
        gw, shards = await self.http.get_shard_count()
        self._gw_url = gw

        return shards

    def guilds_for(self, shard_id: int) -> 'typing.Iterable[dt_guild.Guild]':
        """
        Gets the guilds for this shard.

        :param shard_id: The shard ID to get guilds from.
        :return: A list of :class:`Guild` that client can see on the specified shard.
        """
        return self.state.guilds_for_shard(shard_id)

    def event(self, name: str):
        """
        A convenience decorator to mark a function as an event.

        This will copy it to the events dictionary, where it will be used as an event later on.

        .. code-block:: python3
        
            @bot.event("message_create")
            async def something(ctx, message: Message):
                pass

        :param name: The name of the event.
        """

        def _inner(func):
            f = ev_dec(name)(func)
            self.events.add_event(func=f)
            return func

        return _inner

    # rip in peace old fire_event
    # 2016-2017
    # broke my pycharm
    async def fire_event(self, event_name: str, *args, **kwargs):
        """
        Fires an event.

        This actually passes the arguments to :meth:`.EventManager.fire_event`.
        """
        gateway = kwargs.get("gateway")
        if not self.state.is_ready(gateway.gw_state.shard_id):
            if event_name not in self.IGNORE_READY and not event_name.startswith("gateway_"):
                return

        return await self.events.fire_event(event_name, *args, **kwargs, client=self)

    async def wait_for(self, *args, **kwargs) -> typing.Any:
        """
        Shortcut for :meth:`.EventManager.wait_for`.
        """
        return await self.events.wait_for(*args, **kwargs)

    # Gateway functions
    async def change_status(self, game: Game = None, status: Status = Status.ONLINE,
                            afk: bool = False,
                            shard_id: int = 0):
        """
        Changes the bot's current status.

        :param game: The game object to use. None for no game.
        :param status: The new status. Must be a :class:`.Status` object.
        :param afk: Is the bot AFK? Only useful for userbots.
        :param shard_id: The shard to change your status on.
        """

        gateway = self._gateways[shard_id]
        return await gateway.send_status(
            name=game.name if game else None, type_=game.type if game else None,
            url=game.url if game else None,
            status=status.value, afk=afk
        )

    # HTTP Functions
    async def edit_profile(self, *,
                           username: str = None,
                           avatar: bytes = None,
                           password: str = None):
        """
        Edits the profile of this bot.

        The user is **not** edited in-place - instead, you must wait for the `USER_UPDATE` event to
        be fired on the websocket.

        :param username: The new username of the bot.
        :param avatar: The bytes-like object that represents the new avatar you wish to use.
        :param password: The password to use. Only for user accounts.
        """
        if not self.user.bot and password is None:
            raise ValueError("Password must be passed for user bots")

        if username:
            if any(x in username for x in ('@', ':', '```')):
                raise ValueError("Username must not contain banned characters")

            if username in ("discordtag", "everyone", "here"):
                raise ValueError("Username cannot be a banned username")

            if not 2 <= len(username) <= 32:
                raise ValueError("Username must be 2-32 characters")

        if avatar:
            avatar = base64ify(avatar)

        await self.http.edit_user(username, avatar, password)

    async def edit_avatar(self, path: str):
        """
        A higher-level way to change your avatar.
        This allows you to provide a path to the avatar file instead of having to read it in 
        manually.

        :param path: The path-like object to the avatar file.
        """
        with open(path, 'rb') as f:
            return await self.edit_profile(avatar=f.read())

    async def get_user(self, user_id: int) -> User:
        """
        Gets a user by ID.

        :param user_id: The ID of the user to get.
        :return: A new :class:`.User` object.
        """
        try:
            return self.state._users[user_id]
        except KeyError:
            u = self.state.make_user(await self.http.get_user(user_id))
            # decache it if we need to
            self.state._check_decache_user(u.id)
            return u

    async def get_application(self, application_id: int) -> AppInfo:
        """
        Gets an application by ID.

        :param application_id: The client ID of the application to fetch.
        :return: A new :class:`.AppInfo` object corresponding to the application.
        """
        data = await self.http.get_app_info(application_id=application_id)
        appinfo = AppInfo(self, **data)

        return appinfo

    async def get_webhook(self, webhook_id: int) -> Webhook:
        """
        Gets a webhook by ID.

        :param webhook_id: The ID of the webhook to get.
        :return: A new :class:`.Webhook` object.
        """
        return self.state.make_webhook(await self.http.get_webhook(webhook_id))

    async def get_invite(self, invite_code: str, *,
                         with_counts: bool = True) -> Invite:
        """
        Gets an invite by code.

        :param invite_code: The invite code to get.
        :param with_counts: Return the approximate counts for this invite?
        :return: A new :class:`.Invite` object.
        """
        return Invite(self, **(await self.http.get_invite(invite_code, with_counts=with_counts)))

    async def get_widget(self, guild_id: int) -> Widget:
        """
        Gets a widget from a guild.
        
        :param guild_id: The ID of the guild to get the widget of. 
        :return: A :class:`.Widget` object.
        """
        data = await self.http.get_widget_data(guild_id)
        return Widget(self, **data)

    async def clean_content(self, content: str) -> str:
        """
        Cleans the content of a message, using the bot's cache.

        :param content: The content to clean.
        :return: The cleaned up message.
        """
        final = []
        tokens = content.split(" ")
        # o(2n) loop
        for token in tokens:
            # try and find a channel from public channels
            channel_match = CHANNEL_REGEX.match(token)
            if channel_match is not None:
                channel_id = int(channel_match.groups()[0])
                channel = self.state.find_channel(channel_id)
                if channel is None or channel.type not in \
                        [dt_channel.ChannelType.TEXT, dt_channel.ChannelType.VOICE]:
                    final.append("#deleted-channel")
                else:
                    final.append(f"#{channel.name}")

                continue

            user_match = MENTION_REGEX.match(token)
            if user_match is not None:
                found_name = None
                user_id = int(user_match.groups()[0])
                member_or_user = self.state.find_member_or_user(user_id)
                if member_or_user:
                    found_name = member_or_user.name

                if found_name is None:
                    final.append(token)
                else:
                    final.append(f"@{found_name}")

                continue

            emoji_match = EMOJI_REGEX.match(token)
            if emoji_match is not None:
                final.append(f":{emoji_match.groups()[0]}:")
                continue

            # if we got here, matching failed
            # so just add the token
            final.append(token)

        return " ".join(final)

    # download_ methods
    async def download_guild_member(self, guild_id: int, member_id: int) -> 'dt_member.Member':
        """
        Downloads a :class:`.Member` over HTTP.
        
        .. warning::
            
            The :attr:`.Member.roles` and similar fields will be empty when downloading a Member,
            unless the guild was in cache.
        
        :param guild_id: The ID of the guild which the member is in. 
        :param member_id: The ID of the member to get.
        :return: The :class:`.Member` object downloaded.
        """
        member_data = await self.http.get_guild_member(guild_id=guild_id, member_id=member_id)
        member = dt_member.Member(self, **member_data)
        # this is enough to pick up the cache
        member.guild_id = guild_id

        # manual refcounts :ancap:
        self.state._check_decache_user(member.id)

        return member

    async def download_guild_members(self, guild_id: int, *,
                                     after: int = None, limit: int = 1000,
                                     get_all: bool = True) -> 'typing.Iterable[dt_member.Member]':
        """
        Downloads the members for a :class:`.Guild` over HTTP.
        
        .. warning::
        
            This can take a long time on big guilds.
        
        :param guild_id: The ID of the guild to download members for.
        :param after: The member ID after which to get members for.
        :param limit: The maximum number of members to return. By default, this is 1000 members.
        :param get_all: Should *all* members be fetched?
        :return: An iterable of :class:`.Member`.
        """
        member_data = []
        if get_all is True:
            last_id = 0
            while True:
                next_data = await self.http.get_guild_members(guild_id=guild_id, limit=limit,
                                                              after=last_id)
                # no more members to get
                if not next_data:
                    break

                member_data.extend(next_data)
                # if there's less data than limit, we are finished downloading members
                if len(next_data) < limit:
                    break

                last_id = member_data[-1]["user"]["id"]
        else:
            next_data = await self.http.get_guild_members(guild_id=guild_id, limit=limit,
                                                          after=after)
            member_data.extend(next_data)

        # create the member objects
        members = []
        for datum in member_data:
            m = dt_member.Member(self, **datum)
            m.guild_id = guild_id
            members.append(m)

        return members

    async def download_channels(self, guild_id: int) -> 'typing.List[dt_channel.Channel]':
        """
        Downloads all the :class:`.Channel` for a Guild.
        
        :param guild_id: The ID of the guild to download channels for. 
        :return: An iterable of :class:`.Channel` objects.
        """
        channel_data = await self.http.get_guild_channels(guild_id=guild_id)
        channels = []
        for datum in channel_data:
            channel = dt_channel.Channel(self, **datum)
            channel.guild_id = guild_id
            channels.append(channel)

        return channels

    async def download_guild(self, guild_id: int, *,
                             full: bool = False) -> 'dt_guild.Guild':
        """
        Downloads a :class:`.Guild` over HTTP.
        
        .. warning::
        
            If ``full`` is True, this will fetch and fill ALL objects of the guild, including 
            channels and members. This can take a *long* time if the guild is large.
        
        :param guild_id: The ID of the Guild object to download. 
        :param full: If all extra data should be downloaded alongside it.
        :return: The :class:`.Guild` object downloaded.
        """
        guild_data = await self.http.get_guild(guild_id)
        # create the new guild using the data specified
        guild = dt_guild.Guild(self, **guild_data)
        guild.unavailable = False

        # update the guild store
        self.state._guilds[guild_id] = guild

        if full:
            # download all of the members
            members = await self.download_guild_members(guild_id=guild_id, get_all=True)
            # update the `_members` dict
            guild._members = {m.id: m for m in members}

            # download all of the channels
            channels = await self.download_channels(guild_id=guild_id)
            guild._channels = {c.id: c for c in channels}

        return guild

    @ev_dec(name="gateway_dispatch_received")
    async def handle_dispatches(self, ctx: EventContext, name: str, dispatch: dict):
        """
        Handles dispatches for the client.
        """
        try:
            handle_name = name.lower()
            handler = getattr(self.state, f"handle_{handle_name}")
        except AttributeError:
            logger.warning(f"Got unknown dispatch {name}")
            return
        else:
            logger.debug(f"Processing event {name}")

        try:
            result = handler(ctx.gateway, dispatch)

            if inspect.isawaitable(result):
                result = await result
            elif inspect.isasyncgen(result):
                async with multio.asynclib.finalize_agen(result) as gen:
                    async for i in gen:
                        await self.events.fire_event(i[0], *i[1:], gateway=ctx.gateway, client=self)

                # no more processing after the async gen
                return

            if not isinstance(result, tuple):
                await self.events.fire_event(result, gateway=ctx.gateway, client=self)
            else:
                await self.events.fire_event(result[0], *result[1:], gateway=ctx.gateway,
                                             client=self)

        except Exception:
            logger.exception(f"Error decoding event {name} with data {dispatch}!")
            await self.kill()
            raise

    @ev_dec(name="ready")
    async def handle_ready(self, ctx: 'EventContext'):
        """
        Handles a READY event, dispatching a ``shards_ready`` event when all shards are ready.
        """
        self._ready_state[ctx.shard_id] = True

        if not all(self._ready_state.values()):
            return

        await self.events.fire_event("shards_ready", gateway=self._gateways[ctx.shard_id],
                                     client=self)

    async def handle_shard(self, shard_id: int, shard_count: int):
        """
        Handles a shard.

        :param shard_id: The shard ID to boot and handle.
        :param shard_count: The shard count to send in the identify packet.
        """
        # consume events
        async with open_websocket(self._token, self._gw_url,
                                  shard_id=shard_id, shard_count=shard_count) as gw:
            self._gateways[shard_id] = gw

            try:
                async with multio.asynclib.finalize_agen(gw.events()) as agen:
                    async for event in agen:
                        await self.fire_event(event[0], *event[1:], gateway=gw)
            except Exception as e:  # kill the bot if we failed to parse something
                await self.kill()
                raise
            finally:
                self._gateways.pop(shard_id, None)

    async def start(self, shard_count: int):
        """
        Starts the bot.

        :param shard_count: The number of shards to boot.
        """
        if self.bot_type & BotType.BOT:
            self.application_info = AppInfo(self, **(await self.http.get_app_info(None)))

        # update ready state
        for shard_id in range(shard_count):
            self._ready_state[shard_id] = False

        async with multio.asynclib.task_manager() as tg:
            self.task_manager = tg
            self.events.task_manager = tg

            for shard_id in range(0, shard_count):
                await multio.asynclib.spawn(tg, self.handle_shard, shard_id, shard_count)

    async def run_async(self, *, shard_count: int = 1, autoshard: bool = True):
        """
        Runs the client asynchronously.

        :param shard_count: The number of shards to boot.
        :param autoshard: If the bot should be autosharded.
        """
        if autoshard:
            shard_count = await self.get_shard_count()

        self.shard_count = shard_count
        return await self.start(shard_count)

    async def kill(self) -> None:
        """
        Kills the bot by closing all shards.
        """
        for gateway in self._gateways.values():
            await gateway.close(code=1006, reason="Bot killed", reconnect=False)

    def run(self, *, shard_count: int = 1, autoshard: bool = True, **kwargs):
        """
        Convenience method to run the bot with multio.

        :param shard_count: The number of shards to use. Ignored if autoshard is True.
        :param autoshard: If the bot should be autosharded.
        """

        p = functools.partial(self.run_async, shard_count=shard_count, autoshard=autoshard)
        multio.run(p, **kwargs)

    @classmethod
    def from_token(cls, token: str = None):
        """
        Starts a bot from a token object.

        :param token: The token to use for the bot.
        """
        bot = cls(token)
        return bot.run()



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Represents a Discord IPC packet.

.. currentmodule:: curious.client.packet
"""
import enum
import json
import struct
import uuid
from io import BytesIO

from curio.io import Socket


class IPCOpcode(enum.IntEnum):
    """
    Represents an IPC opcode.
    """
    HANDSHAKE = 0
    FRAME = 1
    CLOSE = 2
    PING = 3
    PONG = 4


class IPCPacket(object):
    """
    Represents an IPC packet.
    """
    def __init__(self, opcode: IPCOpcode, data: dict):
        """
        :param opcode: The :class:`.IPCOpcode` for this packet.
        :param data: A dict of data enclosed in this packet.
        """
        self.opcode = opcode
        self._json_data = data

    @staticmethod
    def _pack_json(data: dict) -> str:
        """
        Packs JSON in a compact representation.
        :param data: The data to pack.
        """
        return json.dumps(data, indent=None, separators=(',', ':'))

    # properties
    @property
    def event(self) -> str:
        """
        Gets the event for this packet. Received packets only.
        """
        return self._json_data["evt"]

    @property
    def cmd(self) -> str:
        """
        Gets the command for this packet.
        """
        return self._json_data["cmd"]

    @property
    def nonce(self) -> uuid.UUID:
        """
        Gets the nonce for this packet.
        """
        return uuid.UUID(self._json_data["nonce"])

    @property
    def data(self) -> str:
        """
        Gets the inner data for this packet.
        """
        return self._json_data["data"]

    def serialize(self) -> bytes:
        """
        Serializes this packet into a series of bytes.
        """
        buf = BytesIO()
        # Add opcode - little endian (why not network order?)
        buf.write(self.opcode.to_bytes(4, byteorder="little"))
        data = self._pack_json(self._json_data)
        # Add data length - little endian (why not network order?)
        buf.write(len(data).to_bytes(4, byteorder="little"))
        # Add data - string, obviously
        buf.write(data.encode("utf-8"))
        return buf.getvalue()

    @classmethod
    def deserialize(cls, data: bytes):
        """
        Deserializes a full packet.

        This method is not usually what you want.
        """
        opcode, length = struct.unpack("<ii", data[:8])
        raw_data = data[8:].decode("utf-8")

        if len(raw_data) != length:
            raise ValueError("Got invalid length.")

        return IPCPacket(IPCOpcode(opcode), json.loads(raw_data))

    @classmethod
    async def read_packet(cls, sock: Socket) -> 'IPCPacket':
        """
        Reads a packet off of the socket, and deserializes it.
        """
        data = await sock.recv(8)

        if len(data) != 8:
            raise ValueError("Got bad IPC read")

        # unpack header so we can get the length
        opcode, length = struct.unpack("<ii", data)

        # read body based on header
        body = await sock.recv(length)
        body_data = json.loads(body.decode("utf-8"))
        return IPCPacket(IPCOpcode(opcode), body_data)



"""
Utilities for ratelimiting a command.
"""
import time
from typing import Any, Callable, List, Tuple

from curious.commands import Context
from curious.commands.exc import CommandRateLimited


class BucketNamer:
    """
    A simple namespace for storing bucket functions.
    """

    def __new__(cls):
        raise NotImplementedError("Don't make an instance of this class")

    @staticmethod
    def GUILD(ctx: Context) -> str:
        """
        A bucket namer that uses the guild ID as the bucket.
        """
        return str(ctx.guild.id)

    @staticmethod
    def CHANNEL(ctx: Context) -> str:
        """
        A bucket namer that uses the channel ID as the bucket.
        """
        return str(ctx.channel.id)

    @staticmethod
    def AUTHOR(ctx: Context) -> str:
        """
        A bucket namer that uses the author ID as the bucket.
        """
        return str(ctx.author.id)

    @staticmethod
    def GLOBAL(ctx: Context) -> str:
        """
        A bucket namer that is global.
        """
        return "GLOBAL"


class CommandRateLimit(object):
    """
    Represents a ratelimit for a command.
    """

    def __init__(self, *, limit: int, time: float,
                 bucket_namer: Callable[[Context], str] = BucketNamer.AUTHOR):
        """
        :param limit: The number of times a command can be called in the specified limit.
        :param time: The time (in seconds) this ratelimit lasts.
        :param bucket_namer: A callable that gets the ratelimit bucket name.
        """
        self.limit = limit
        self.time = time
        self.bucket_namer = bucket_namer

        #: The command function being used.
        self.command = None

    def get_full_bucket_key(self, ctx: Context) -> Tuple[str, str]:
        """
        Gets the full bucket key for this ratelimit.
        """
        return self.command.cmd_name, self.bucket_namer(ctx)


class RateLimiter(object):
    """
    Represents a ratelimiter. This ensures that commands meet the ratelimit before being ran.
    """

    def __init__(self):
        self._ratelimit_buckets = {}

    async def update_bucket(self, key: Any, current_uses: int, expiration: float):
        """
        Updates a ratelimit bucket.

        :param key: The ratelimit key to use.
        :param current_uses: The current uses for the key.
        :param expiration: When the ratelimit expires.
        """
        self._ratelimit_buckets[key] = (current_uses, expiration)

    async def get_bucket(self, key: Any) -> Tuple[int, float]:
        """
        Gets the ratelimit bucket for the specified key.

        :param key: The key to use.
        :return: A two-item tuple of (uses, expiration), or None if no bucket was found.
        """
        return self._ratelimit_buckets.get(key)

    async def ensure_ratelimits(self, ctx: Context, cmd):
        """
        Ensures the ratelimits for a command.
        """
        ratelimits: List[CommandRateLimit] = cmd.cmd_ratelimits
        for limit in ratelimits:
            bucket_key = limit.get_full_bucket_key(ctx)
            bucket = await self.get_bucket(key=bucket_key)
            if not bucket:
                await self.update_bucket(bucket_key, 1, time.monotonic() + limit.time)
                continue

            # check if we've hit the limit
            if bucket[0] == limit.limit:
                # we have, but that might be okay if the timer has expired
                if time.monotonic() > bucket[1]:
                    # we're good, so we can just reset the bucket and continue on our way
                    await self.update_bucket(bucket_key, 1, time.monotonic() + limit.time)
                else:
                    raise CommandRateLimited(ctx, cmd, limit, bucket)
            else:
                # we haven't, but we need to up the number anyway
                await self.update_bucket(bucket_key, bucket[0] + 1, bucket[1])



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Invite objects.

Invite objects are linked to a real channel and real guild, but Discord does not return the full
data for these objects. Therefore, a few "mini" objects are provided that represent the objects:

 - :class:`.InviteGuild`, which contains the ``name``, ``splash_url`` and ``icon_url`` of the
   Guild.
 - :class:`.InviteChannel`, which contains the ``name`` and ``type`` of the Channel.
 
These objects will be returned on :attr`.Invite.guild` and :attr:`.Invite.channel` respectively
if the data for each  is not cached by curious. Otherwise, full :class:`.Guild` and
:class:`.Channel` objects will be returned.

.. currentmodule:: curious.dataclasses.invite
"""
import typing

from curious import util
from curious.dataclasses import channel as dt_channel, guild as dt_guild, member as dt_member, \
    user as dt_user
from curious.dataclasses.bases import IDObject
from curious.exc import PermissionsError


class InviteGuild(IDObject):
    """
    Represents an InviteGuild - a subset of a guild.
    """

    def __init__(self, **kwargs):
        super().__init__(kwargs.get("id"))

        #: The name of this guild.
        self.name = kwargs.get("name", None)  # type: str

        #: The splash hash of this guild.
        self.splash_hash = kwargs.get("splash")  # type: str

        #: The icon hash of this guild.
        self._icon_hash = kwargs.get("icon")  # type: str

        #: A list of features for this guild.
        self.features = kwargs.get("features")  # type: typing.List[str]

        #: The approximate member count for this guild.
        self.member_count = kwargs.get("approximate_member_count")

        #: The approximate presence count.
        self.presence_count = kwargs.get("approximate_presence_count")

        #: The number of text channels.
        self.text_channel_count = kwargs.get("text_channel_count")

        #: The number of voice channels.
        self.voice_channel_count = kwargs.get("voice_channel_count")

    def __repr__(self) -> str:
        return "<InviteGuild id={} name='{}'>".format(self.id, self.name)

    __str__ = __repr__

    @property
    def icon_url(self) -> str:
        """
        :return: The icon URL for this guild, or None if one isn't set.
        """
        if self._icon_hash:
            return "https://cdn.discordapp.com/icons/{}/{}.webp".format(self.id, self._icon_hash)

    @property
    def splash_url(self) -> str:
        """
        :return: The splash URL for this guild, or None if one isn't set.
        """
        if self.splash_hash:
            return "https://cdn.discordapp.com/splashes/{}/{}.webp".format(self.id,
                                                                           self.splash_hash)


class InviteChannel(IDObject):
    """
    Represents an InviteChannel - a subset of a channel.
    """

    def __init__(self, **kwargs):
        super().__init__(kwargs.get("id"))

        #: The name of this channel.
        self.name = kwargs.get("name")

        #: The :class:`.ChannelType` of this channel.
        self.type = dt_channel.ChannelType(kwargs.get("type"))

    def __repr__(self) -> str:
        return "<InviteChannel name={}>".format(self.name)


class InviteMetadata(object):
    """
    Represents metadata attached to an invite.
    """

    __slots__ = "uses", "max_uses", "max_age", "temporary", "created_at", "revoked",

    def __init__(self, **kwargs):
        #: The number of times this invite was used.
        self.uses = kwargs.get("uses", 0)  # type: int

        #: The maximum number of uses this invite can use.
        self.max_uses = kwargs.get("max_uses", 0)  # type: int

        #: The maximum age of this invite.
        self.max_age = kwargs.get("max_age", 0)  # type: int

        #: Is this invite temporary?
        self.temporary = kwargs.get("temporary", False)  # type: bool

        #: When was this invite created at?
        self.created_at = util.to_datetime(kwargs.get("created_at", None))

        #: Is this invite revoked?
        self.revoked = kwargs.get("revoked", False)  # type: bool


class Invite(object):
    """
    Represents an invite object.
    """

    def __init__(self, client, **kwargs):
        self._bot = client

        #: The invite code.
        self.code = kwargs.get("code")  # type: str

        #: The guild ID for this invite.
        self.guild_id = int(kwargs["guild"]["id"])

        #: The channel ID for this invite.
        self.channel_id = int(kwargs["channel"]["id"])

        #: The invite guild this is attached to.
        #: The actual guild object can be more easily fetched with `.guild`.
        self._invite_guild = \
            InviteGuild(**kwargs.get("guild"))  # type: typing.Union[InviteGuild, dt_guild.Guild]

        #: The invite channel this is attached to.
        #: The actual channel object can be more easily fetched with `.channel`.
        self._invite_channel = \
            InviteChannel(**kwargs.get("channel"))  # type: typing.Union[InviteChannel, dt_channel.Channel]

        #: The ID of the user that created this invite.
        #: This can be None for partnered invites.
        self.inviter_id = None  # type: int

        if "inviter" in kwargs:
            self._inviter_data = kwargs["inviter"]
            self.inviter_id = int(self._inviter_data.get("id", 0))

        #: The invite metadata object associated with this invite.
        #: This can be None if the invite has no metadata.
        if "uses" not in kwargs:
            self._invite_metadata = None
        else:
            self._invite_metadata = InviteMetadata(**kwargs)

    def __repr__(self) -> str:
        return "<Invite code={} guild={} channel={}>".format(self.code, self.guild, self.channel)

    def __del__(self) -> None:
        if self.inviter:
            self._bot.state._check_decache_user(self.inviter_id)

    @property
    def inviter(self) -> 'typing.Union[dt_member.Member, dt_user.User]':
        """
        :return: The :class:`.Member` or :class:`.User` that made this invite.
        """
        if isinstance(self._invite_guild, InviteGuild):
            return self._bot.state.make_user(self._inviter_data)

        u = self._invite_guild.members.get(self.inviter_id)
        if not u:
            return self._bot.state.make_user(self._inviter_data)

        return u

    @property
    def guild(self) -> 'typing.Union[dt_guild.Guild, InviteGuild]':
        """
        :return: The guild this invite is associated with.
        """
        return self._bot.state.guilds.get(self.guild_id, self._invite_guild)

    @property
    def channel(self) -> 'typing.Union[dt_channel.Channel, InviteChannel]':
        """
        :return: The channel this invite is associated with.
        """
        g = self.guild
        if g == self._invite_guild:
            return self._invite_channel

        return g.channels.get(self.channel_id, self._invite_channel)

    async def delete(self):
        """
        Deletes this invite.

        You must have MANAGE_CHANNELS permission in the guild to delete the invite.
        """
        guild = self.guild
        if guild != self._invite_guild:
            if not guild.me.guild_permissions.manage_channels:
                raise PermissionsError("manage_channels")

        await self._bot.http.delete_invite(self.code)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Message objects.

.. currentmodule:: curious.dataclasses.message
"""
import enum
import re
import typing

from curious.dataclasses import channel as dt_channel, emoji as dt_emoji, guild as dt_guild, \
    invite as dt_invite, member as dt_member, role as dt_role, user as dt_user, \
    webhook as dt_webhook
from curious.dataclasses.attachment import Attachment
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.embed import Embed
from curious.exc import CuriousError, ErrorCode, HTTPException, PermissionsError
from curious.util import AsyncIteratorWrapper, to_datetime

CHANNEL_REGEX = re.compile(r"<#([0-9]*)>")
INVITE_REGEX = re.compile(r"(?:discord\.gg/(\S+)|discordapp\.com/invites/(\S+))")
EMOJI_REGEX = re.compile(r"<a?:([\S]+):([0-9]+)>")
MENTION_REGEX = re.compile(r"<@!?([0-9]+)>")


class MessageType(enum.IntEnum):
    """
    Represents the type of a message.
    """
    #: The default (i.e. user message) type.
    DEFAULT = 0

    # 1 through 5 are groups only
    #: The recipient add type, used when a recipient is added to a group.
    RECIPIENT_ADD = 1

    #: The recipient remove type, used when a recipient is added to a group.
    RECIPIENT_REMOVE = 2

    #: The call type, used when a call is started.
    CALL = 3

    #: The channel name change type, used when a group channel name is changed.
    CHANNEL_NAME_CHANGE = 4

    #: The channel icon change type, used when a group channel icon is changed.
    CHANNEL_ICON_CHANGE = 5

    #: The channel pinned message type, used when a message is pinned.
    CHANNEL_PINNED_MESSAGE = 6

    #: The guild member join type, used when a member joins a guild.
    GUILD_MEMBER_JOIN = 7


class Message(Dataclass):
    """
    Represents a Message.
    """
    __slots__ = ("content", "guild_id", "author", "created_at", "edited_at", "embeds",
                 "attachments", "_mentions", "_role_mentions", "reactions", "channel_id",
                 "author_id", "type")

    def __init__(self, client, **kwargs):
        super().__init__(kwargs.get("id"), client)

        #: The content of the message.
        self.content = kwargs.get("content", None)  # type: str

        #: The ID of the guild this message is in.
        self.guild_id = None

        #: The ID of the channel the message was sent in.
        self.channel_id = int(kwargs.get("channel_id", 0))  # type: int

        #: The ID of the author.
        self.author_id = int(kwargs.get("author", {}).get("id", 0)) or None  # type: int

        #: The author of this message. Can be one of: :class:`.Member`, :class:`.Webhook`,
        #: :class:`.User`.
        self.author = None  # type: typing.Union[dt_member.Member, dt_webhook.Webhook]

        type_ = kwargs.get("type", 0)
        #: The type of this message.
        self.type = MessageType(type_)

        #: The true timestamp of this message, a :class:`datetime.datetime`.
        #: This is not the snowflake timestamp.
        self.created_at = to_datetime(kwargs.get("timestamp", None))

        #: The edited timestamp of this message.
        #: This can sometimes be None.
        edited_timestamp = kwargs.get("edited_timestamp", None)
        if edited_timestamp is not None:
            self.edited_at = to_datetime(edited_timestamp)
        else:
            self.edited_at = None

        #: The list of :class:`.Embed` objects this message contains.
        self.embeds = []
        for embed in kwargs.get("embeds", []):
            self.embeds.append(Embed(**embed))

        #: The list of :class:`.Attachment` this message contains.
        self.attachments = []

        for attachment in kwargs.get("attachments", []):
            self.attachments.append(Attachment(bot=self._bot, **attachment))

        #: The mentions for this message.
        #: This is UNORDERED.
        self._mentions = kwargs.get("mentions", [])

        #: The role mentions for this message.
        #: This is UNORDERED.
        self._role_mentions = kwargs.get("mention_roles", [])

        #: The reactions for this message.
        self.reactions = []

    def __repr__(self) -> str:
        return "<{0.__class__.__name__} id={0.id} content='{0.content}'>".format(self)

    def __str__(self) -> str:
        return self.content

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` this message is associated with.
        """
        return self.channel.guild

    @property
    def channel(self) -> 'dt_channel.Channel':
        """
        :return: The :class:`.Channel` this message is associated with.
        """
        return self._bot.state.find_channel(self.channel_id)

    @property
    def mentions(self) -> 'typing.List[dt_member.Member]':
        """
        Returns a list of :class:`.Member` that were mentioned in this message.
        
        .. warning::
            
            The mentions in this will **not** be in order. Discord does not return them in any 
            particular order.

        """
        return self._resolve_mentions(self._mentions, "member")

    @property
    def role_mentions(self) -> 'typing.List[dt_role.Role]':
        """
        Returns a list of :class:`.Role` that were mentioned in this message.
        
        .. warning::
            
            The mentions in this will **not** be in order. Discord does not return them in any 
            particular order.

        """

        return self._resolve_mentions(self._role_mentions, "role")

    @property
    def channel_mentions(self) -> 'typing.List[dt_channel.Channel]':
        """
        Returns a list of :class:`.Channel` that were mentioned in this message.
        
        .. note::
        
            These mentions **are** in order. They are parsed from the message content.

        """
        mentions = CHANNEL_REGEX.findall(self.content)
        return self._resolve_mentions(mentions, "channel")

    @property
    def emojis(self) -> 'typing.List[dt_emoji.Emoji]':
        """
        Returns a list of :class:`.Emoji` that was found in this message.
        """
        matches = EMOJI_REGEX.findall(self.content)
        emojis = []

        for (name, i) in matches:
            e = self.guild.emojis.get(int(i))
            if e:
                emojis.append(e)

        return emojis

    async def clean_content(self) -> str:
        """
        Gets the cleaned content for this message.
        """
        return await self._bot.clean_content(self.content)

    async def get_invites(self) -> 'typing.List[dt_invite.Invite]':
        """
        Gets a list of valid invites in this message.
        """
        invites = INVITE_REGEX.findall(self.content)
        obbs = []
        for match in invites:
            if match[0]:
                code = match[0]
            else:
                code = match[1]

            try:
                obbs.append(await self._bot.get_invite(code))
            except HTTPException as e:
                if e.error_code != ErrorCode.UNKNOWN_INVITE:
                    raise

        return obbs

    @property
    def invites(self) -> 'typing.AsyncIterator[dt_invite.Invite]':
        """
        Returns a list of :class:`.Invite` objects that are in this message (and valid).
        """
        return AsyncIteratorWrapper(self.get_invites)

    def _resolve_mentions(self,
                          mentions: typing.List[typing.Union[dict, str]],
                          type_: str) \
            -> 'typing.List[typing.Union[dt_channel.Channel, dt_role.Role, dt_member.Member]]':
        """
        Resolves the mentions for this message.
        
        :param mentions: The mentions to resolve; a list of dicts or ints.
        :param type_: The type of mention to resolve: ``channel``, ``role``, or ``member``.
        """
        final_mentions = []
        for mention in mentions:
            obb = None
            if type_ == "member":
                id = int(mention["id"])
                obb = self.guild.members.get(id)
                if obb is None:
                    obb = self._bot.state.make_user(mention)
                    # always check for a decache
                    self._bot.state._check_decache_user(id)
            elif type_ == "role":
                obb = self.guild.roles.get(int(mention))
            elif type_ == "channel":
                obb = self.guild.channels.get(int(mention))

            if obb is not None:
                final_mentions.append(obb)

        return final_mentions

    def reacted(self, emoji: 'typing.Union[dt_emoji.Emoji, str]') -> bool:
        """
        Checks if this message was reacted to with the specified emoji.

        :param emoji: The emoji to check.
        """
        for reaction in self.reactions:
            if reaction.emoji == emoji:
                return True

        return False

    # Message methods
    async def delete(self) -> None:
        """
        Deletes this message.

        You must have MANAGE_MESSAGE permissions to delete this message, or have it be your own 
        message.
        """
        if self.guild is None:
            me = self._bot.user.id
            has_manage_messages = False
        else:
            me = self.guild.me.id
            has_manage_messages = self.channel.permissions(self.guild.me).manage_messages

        if self.id != me and not has_manage_messages:
            raise PermissionsError("manage_messages")

        await self._bot.http.delete_message(self.channel.id, self.id)

    async def edit(self, new_content: str = None, *,
                   embed: Embed = None) -> 'Message':
        """
        Edits this message.

        You must be the owner of this message to edit it.

        :param new_content: The new content for this message.
        :param embed: The new embed to provide.
        :return: This message, but edited with the new content.
        """
        if self.guild is None:
            is_me = self.author not in self.channel.recipients
        else:
            is_me = self.guild.me == self.author

        if not is_me:
            raise CuriousError("Cannot edit messages from other users")

        if embed:
            embed = embed.to_dict()

        async with self._bot.events.wait_for_manager("message_update",
                                                     lambda o, n: n.id == self.id):
            await self._bot.http.edit_message(self.channel.id, self.id, content=new_content,
                                              embed=embed)
        return self

    async def pin(self) -> 'Message':
        """
        Pins this message.

        You must have MANAGE_MESSAGES in the channel to pin the message.
        """
        if self.guild is not None:
            if not self.channel.permissions(self.guild.me).manage_messages:
                raise PermissionsError("manage_messages")

        await self._bot.http.pin_message(self.channel.id, self.id)
        return self

    async def unpin(self) -> 'Message':
        """
        Unpins this message.

        You must have MANAGE_MESSAGES in this channel to unpin the message.
        Additionally, the message must already be pinned.
        """
        if self.guild is not None:
            if not self.channel.permissions(self.guild.me).manage_messages:
                raise PermissionsError("manage_messages")

        await self._bot.http.unpin_message(self.channel.id, self.id)
        return self

    async def get_who_reacted(self, emoji: 'typing.Union[dt_emoji.Emoji, str]') \
            -> 'typing.List[typing.Union[dt_user.User, dt_member.Member]]':
        """
        Fetches who reacted to this message.

        :param emoji: The emoji to check.
        :return: A list of either :class:`.Member` or :class:`.User` that reacted to this message.
        """
        if isinstance(emoji, dt_emoji.Emoji):
            emoji = "{}:{}".format(emoji.name, emoji.id)

        reactions = await self._bot.http.get_reaction_users(self.channel.id, self.id, emoji)
        result = []

        for user in reactions:
            member_id = int(user.get("id"))
            if self.guild is None:
                result.append(dt_user.User(self._bot, **user))
            else:
                member = self.guild.members.get(member_id)
                if not member:
                    result.append(dt_user.User(self._bot, **user))
                else:
                    result.append(member)

        return result

    async def react(self, emoji: 'typing.Union[dt_emoji.Emoji, str]'):
        """
        Reacts to a message with an emoji.

        This requires an Emoji object for reacting to messages with custom reactions, or a string 
        containing the literal unicode (e.g ™) for normal emoji reactions.

        :param emoji: The emoji to react with.
        """
        if self.guild:
            if not self.channel.permissions(self.guild.me).add_reactions:
                # we can still add already reacted emojis
                # so make sure to check for that
                if not self.reacted(emoji):
                    raise PermissionsError("add_reactions")

        if isinstance(emoji, dt_emoji.Emoji):
            # undocumented!
            emoji = "{}:{}".format(emoji.name, emoji.id)

        await self._bot.http.add_reaction(self.channel.id, self.id, emoji)

    async def unreact(self, reaction: 'typing.Union[dt_emoji.Emoji, str]',
                      victim: 'dt_member.Member' = None):
        """
        Removes a reaction from a user.

        :param reaction: The reaction to remove.
        :param victim: The victim to remove the reaction of. Can be None to signify ourselves.
        """
        if not self.guild:
            if victim and victim != self:
                raise CuriousError("Cannot delete other reactions in a DM")

        if victim and victim != self:
            if not self.channel.permissions(self.guild.me).manage_messages:
                raise PermissionsError("manage_messages")

        if isinstance(reaction, dt_emoji.Emoji):
            emoji = "{}:{}".format(reaction.name, reaction.id)
        else:
            emoji = reaction

        await self._bot.http.delete_reaction(self.channel.id, self.id, emoji,
                                             victim=victim.id if victim else None)

    async def remove_all_reactions(self) -> None:
        """
        Removes all reactions from a message.
        """
        if not self.guild:
            raise CuriousError("Cannot delete other reactions in a DM")

        if not self.channel.permissions(self.guild.me).manage_messages:
            raise PermissionsError("manage_messages")

        await self._bot.http.delete_all_reactions(self.channel.id, self.id)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains the class for the commands manager for a client.

.. currentmodule:: curious.commands.manager
"""
import importlib
import inspect
import logging
import sys
import traceback
import typing
from collections import defaultdict
from functools import partial

import multio

from curious.commands.context import Context
from curious.commands.exc import CommandsError
from curious.commands.help import help_command
from curious.commands.plugin import Plugin
from curious.commands.ratelimit import RateLimiter
from curious.commands.utils import prefix_check_factory
from curious.core import client as md_client
from curious.core.event import EventContext, event
from curious.dataclasses.message import Message

logger = logging.getLogger("curious.commands.manager")


class CommandsManager(object):
    """
    A manager that handles commands for a client.

    First, you need to create the manager and attach it to a client:

    .. code-block:: python3

        # form 1, automatically register with the client
        manager = CommandsManager.with_client(bot)

        # form 2, manually register
        manager = CommandsManager(bot)
        manager.register_events()

    This is required to add the handler events to the client.

    Next, you need to register a message check handler. This is a callable that is called for
    every message to try and extract the command from a message, if it matches.
    By default, the manager provides an easy way to use a simple command prefix:

    .. code-block:: python3

        # at creation time
        manager = CommandsManager(bot, command_prefix="!")

        # or set it on the manager
        manager.command_prefix = "!"

    At this point, the command prefix will be available on the manager with either
    :attr:`.Manager.command_prefix` or :attr:`.Manager.message_check.prefix`.

    If you need more complex message checking, you can use ``message_check``:

    .. code-block:: python3

        manager = CommandsManager(bot, message_check=my_message_checker)
        # or
        manager.message_check = my_message_checker

    Finally, you can register plugins or modules containing plugins with the manager:

    .. code-block:: python3

        @bot.event("ready")
        async def load_plugins(ctx: EventContext):
            # load plugin explicitly
            await manager.load_plugin(PluginClass, arg1)
            # load plugins from a module
            await manager.load_plugins_from("my.plugin.module")

    You can also add free-standing commands that aren't bound to a plugin with
    :meth:`.CommandsManager.add_command`:

    .. code-block:: python3

        @command()
        async def ping(ctx: CommandsContext):
            await ctx.channel.messages.send(content="Ping!")

        manager.add_command(ping)

    These will then be available to the client.
    """

    def __init__(self, client: 'md_client.Client', *,
                 message_check=None, command_prefix: str = None):
        """
        :param client: The :class:`.Client` to use with this manager.
        :param message_check: The message check function for this manager.

            This should take two arguments, the client and message, and should return either None
            or a 2-item tuple:
              - The command word matched
              - The tokens after the command word
        """
        if message_check is None and command_prefix is None:
            raise ValueError("Must provide one of message_check or command_prefix")

        #: The client for this manager.
        self.client = client

        if message_check is None:
            message_check = prefix_check_factory(command_prefix)

        #: The message check function for this manager.
        self.message_check = message_check

        #: A dictionary mapping of <plugin name> -> <plugin> object.
        self.plugins = {}

        #: A dictionary of stand-alone commands, i.e. commands not associated with a plugin.
        self.commands = {}

        #: The current ratelimiter.
        self.ratelimiter = RateLimiter()

        self._module_plugins = defaultdict(lambda: [])

    @classmethod
    def with_client(cls, client: 'md_client.Client', **kwargs):
        """
        Creates a manager and automatically registers events.
        """
        obb = cls(client=client, **kwargs)
        obb.register_events()
        return obb

    def register_events(self) -> None:
        """
        Copies the events to the client specified on this manager.
        """
        self.client.events.add_event(self.handle_message)
        self.client.events.add_event(self.default_command_error)
        self.client.events.add_event_hook(self.event_hook)

        from curious.commands.decorators import command
        self.commands["help"] = command(name="help")(help_command)

    async def load_plugin(self, klass: typing.Type[Plugin], *args,
                          module: str = None):
        """
        Loads a plugin.

        .. note::

            The client instance will automatically be provided to the Plugin's ``__init__``.

        :param klass: The plugin class to load.
        :param args: Any args to provide to the plugin.
        :param module: The module name provided with this plugin. Only used interally.
        """
        # get the name and create the plugin object
        plugin_name = getattr(klass, "plugin_name", klass.__name__)
        instance = klass(self.client, *args)

        # call load, of course
        await instance.load()

        self.plugins[plugin_name] = instance
        if module is not None:
            self._module_plugins[module].append(instance)

        return instance

    async def unload_plugin(self, klass: typing.Union[Plugin, str]):
        """
        Unloads a plugin.

        :param klass: The plugin class or name of plugin to unload.
        """
        p: Plugin = None
        if isinstance(klass, str):
            p = self.plugins.pop(klass)

        for k, p in self.plugins.copy().items():
            if type(p) == klass:
                p = self.plugins.pop(k)
                break

        if p is not None:
            # cancel the task group used for this plugin, if it's running
            if p.task_group is not None:
                await multio.asynclib.cancel_task_group(p.task_group)

            await p.unload()

        return p

    def _lookup_command(self, name: str):
        """
        Does a lookup in plugin and standalone commands.
        """
        if name in self.commands:
            return self.commands[name]

        for plugin in self.plugins.values():
            cmds = plugin._get_commands()

            try:
                return next(filter(lambda cmd: not cmd.cmd_subcommand and
                                   (cmd.cmd_name == name or name in cmd.cmd_aliases), cmds))
            except StopIteration:
                continue

    def get_command(self, command_name: str):
        """
        Gets a command from the internal command storage.
        If provided a string separated by spaces, a subcommand lookup will be attempted.

        :param command_name: The name of the command to lookup.
        """
        # do an immediate lookup for the first token
        sp = command_name.split(" ")
        command = self._lookup_command(sp[0])

        if command is None:
            return None

        for token in sp[1:]:
            try:
                filtered = filter(lambda cmd: cmd.cmd_name == token or token in cmd.cmd_aliases,
                                  command.cmd_subcommands)
                command = next(filtered)
            except StopIteration:
                return None

        return command

    def add_command(self, command):
        """
        Adds a command.

        :param command: A command function.
        """
        if not hasattr(command, "is_cmd"):
            raise ValueError("Commands must be decorated with the command decorator")

        self.commands[command.cmd_name] = command
        return command

    def remove_command(self, command):
        """
        Removes a command.

        :param command: The name of the command, or the command function.
        """
        if isinstance(command, str):
            return self.commands.pop(command)
        else:
            for k, p in self.commands.copy().items():
                if p == command:
                    return self.commands.pop(k)

    async def load_plugins_from(self, import_path: str):
        """
        Loads plugins from the specified module.

        :param import_path: The import path to import.
        """
        mod = importlib.import_module(import_path)

        # define the predicate for the body scanner
        def predicate(item):
            if not isinstance(item, type):
                return False

            # only accept plugin subclasses
            if not issubclass(item, Plugin):
                return False

            # ensure item is not actually Plugin
            if item == Plugin:
                return False

            # it is a plugin
            return True

        for plugin_name, plugin_class in inspect.getmembers(mod, predicate=predicate):
            await self.load_plugin(plugin_class, module=mod)

    async def unload_plugins_from(self, import_path: str):
        """
        Unloads plugins from the specified module.
        This will delete the module from sys.path.

        :param import_path: The import path.
        """
        for plugin in self._module_plugins[import_path]:
            await plugin.unload()
            self.plugins.pop(getattr(plugin, "plugin_name", "__name__"))

        del sys.modules[import_path]
        del self._module_plugins[import_path]

    async def event_hook(self, ctx: EventContext, *args, **kwargs):
        """
        The event hook for the commands manager.
        """
        async with multio.asynclib.task_manager() as tg:
            for plugin in self.plugins.values():
                body = inspect.getmembers(plugin, predicate=lambda v: hasattr(v, "is_event"))
                for _, handler in body:
                    if ctx.event_name not in handler.events:
                        continue

                    cofunc = partial(self.client.events._safety_wrapper,
                                     handler, ctx, *args, **kwargs)

                    await multio.asynclib.spawn(tg, cofunc)

    async def handle_commands(self, ctx: EventContext, message: Message):
        """
        Handles commands for a message.
        """
        # don't process messages pre-cache
        if not message.author:
            return

        # check bot type
        if message.author.user.bot and self.client.bot_type & 8:
            return

        if message.author.user != self.client.user and self.client.bot_type & 64:
            return

        if message.guild_id is not None and self.client.bot_type & 32:
            return

        if message.guild_id is None and self.client.bot_type & 16:
            return

        # step 1, match the messages
        matched = self.message_check(self.client, message)
        if inspect.isawaitable(matched):
            matched = await matched

        if matched is None:
            return None

        # deconstruct the tuple returned into more useful variables than a single tuple
        command_word, tokens = matched

        # step 2, create the new commands context
        ctx = Context(event_context=ctx, message=message)
        ctx.command_name = command_word
        ctx.tokens = tokens
        ctx.manager = self

        # step 3, invoke the context to try and match the command and run it
        await ctx.try_invoke()

    @event("command_error")
    async def default_command_error(self, ev_ctx: EventContext, ctx: Context, err: CommandsError):
        """
        Handles command errors by default.
        """
        # autoremove ourself if applicable
        if len(self.client.events.event_listeners.getall("command_error")) > 1:
            self.client.events.remove_event("command_error", self.default_command_error)
            return

        fmtted = ''.join(traceback.format_exception(type(err), err, err.__traceback__))
        logger.error(f"Error in command!\n{fmtted}")

    @event("message_create")
    async def handle_message(self, ctx: EventContext, message: Message):
        """
        Registered as the event handler in a client for handling commands.
        """
        return await self.handle_commands(ctx, message)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for User objects.

.. currentmodule:: curious.dataclasses.user
"""

import datetime

from curious.dataclasses import channel as dt_channel, guild as dt_guild, message as dt_message
from curious.dataclasses.bases import Dataclass
from curious.exc import CuriousError


class AvatarUrl(object):
    """
    Represents a user's avatar URL.

    To get the actual URL, do str(avatar_url).
    """

    def __init__(self, user: 'User') -> None:
        """
        :param user: The :class:`.User` for this URL.
        """
        self._user = user
        self._format = "webp"
        self._size = 256

    def __str__(self) -> str:
        """
        :return: The string URL for this avatar URL.
        """
        if not self._user.avatar_hash:
            base_url = f"https://cdn.discordapp.com/embed/avatars/" \
                       f"{int(self._user.discriminator) % 5}"
        else:
            base_url = f"https://cdn.discordapp.com/avatars/" \
                       f"{self._user.id}/{self._user.avatar_hash}"

        return f"{base_url}.{self._format}?size={self._size}"

    def as_format(self, format: str) -> 'AvatarUrl':
        """
        Gets the URL in the specified format.

        :param format: The format to use. Usually ``png``, ``webp`` or ``gif``.
        :return: A new :class:`.AvatarUrl` with the specified format.
        """
        obb = AvatarUrl(self._user)
        obb._format = format
        obb._size = self._size
        return obb

    def with_size(self, size: int) -> 'AvatarUrl':
        """
        Gets the URL in the specified size.

        :param size: The size for the URL.
        :return: A new :class:`.AvatarUrl` with the specified size.
        """
        obb = AvatarUrl(self._user)
        obb._format = self._format
        obb._size = size
        return obb

    def __eq__(self, other: 'AvatarUrl'):
        if not isinstance(other, AvatarUrl):
            return NotImplemented

        return str(self) == str(other)

    def __lt__(self, other):
        if not isinstance(other, AvatarUrl):
            return NotImplemented

        return str(self) < str(other)


class User(Dataclass):
    """
    This represents a bare user - i.e, somebody without a guild attached.
    This is used in DMs and similar. All member objects have a reference to their user on ``.user``.

    :ivar id: The ID of this user.
    """

    __slots__ = ("username", "discriminator", "avatar_hash", "verified", "mfa_enabled",
                 "bot", "_bot")

    def __init__(self, client, **kwargs):
        super().__init__(kwargs.get("id"), client)

        #: The username of this user.
        self.username = kwargs.get("username", None)

        #: The discriminator of this user.
        #: Note: This is a string, not an integer.
        self.discriminator = kwargs.get("discriminator", None)

        #: The avatar hash of this user.
        self.avatar_hash = kwargs.get("avatar", None)

        #: If this user is verified or not.
        self.verified = kwargs.get("verified", None)

        #: If this user has MFA enabled or not.
        self.mfa_enabled = kwargs.get("mfa_enabled", None)

        #: If this user is a bot.
        self.bot = kwargs.get("bot", False)

    @property
    def user(self) -> 'User':
        return self

    def _copy(self):
        new_object = object.__new__(self.__class__)
        new_object.id = self.id
        new_object.username = self.username
        new_object.discriminator = self.discriminator
        new_object.avatar_hash = self.avatar_hash
        new_object.verified = self.verified
        new_object.mfa_enabled = self.mfa_enabled
        new_object.bot = self.bot

        new_object._bot = self._bot

        return new_object

    @property
    def avatar_url(self) -> 'AvatarUrl':
        """
        :return: The avatar URL of this user.
        """
        return AvatarUrl(self)

    @property
    def static_avatar_url(self) -> str:
        """
        :return: The avatar URL of this user, but static.
        """
        return str(self.avatar_url.as_format('png'))

    @property
    def name(self) -> str:
        return self.username

    @property
    def mention(self) -> str:
        """
        :return: A string that mentions this user.
        """
        return "<@{}>".format(self.id)

    @property
    def created_at(self) -> datetime.datetime:
        """
        :return: The :class:`.datetime.datetime` this user was created at.
        """
        return self.snowflake_timestamp

    def __repr__(self) -> str:
        return "<{} id={} name={} discrim={}>".format(type(self).__name__, self.id, self.name,
                                                      self.discriminator)

    def __str__(self) -> str:
        return f"{self.username}#{self.discriminator}"

    async def open_private_channel(self) -> 'dt_channel.Channel':
        """
        Opens a private channel with a user.

        .. note::

            You probably won't need this to just send messages to a user.

        :return: The newly created private channel.
        """
        if self.discriminator == "0000":
            raise CuriousError("Cannot open a private channel with a webhook")

        # First, try and access the channel from the channel cache.
        original_channel = self._bot.state.find_channel(self.id)
        if original_channel:
            return original_channel

        # Failing that, open a new private channel.
        channel_data = await self._bot.http.create_private_channel(self.id)
        channel = self._bot.state.make_private_channel(channel_data)
        return channel

    async def send(self, content: str = None, *args, **kwargs) -> 'dt_message.Message':
        """
        Sends a message to the user over a private channel.

        :param content: The contet of the message to send.
        :return: A new :class:`.Message` representing the sent message.
        """
        channel = await self.open_private_channel()
        message = await channel.messages.send(content, *args, **kwargs)

        return message

    async def unban_from(self, guild: 'dt_guild.Guild'):
        """
        Unbans this user from a guild.

        :param guild: The :class:`.Guild` to unban in.
        """
        return await guild.bans.remove(self)


class BotUser(User):
    """
    A special type of user that represents ourselves.
    """

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        #: The email for this user.
        self.email = kwargs.get("email", None)

        #: Does this user use mobile?
        self.mobile = kwargs.get("mobile", False)

        #: Is this user premium?
        self.premium = kwargs.get("premium", False)

    async def open_private_channel(self):
        raise NotImplementedError("Cannot open a private channel with yourself")

    async def send(self, *args, **kwargs):
        raise NotImplementedError("Cannot send messages to your own user")

    async def edit(self, *args, **kwargs):
        """
        Edits the bot's current profile.
        """
        return await self._bot.edit_profile(*args, **kwargs)

    async def upload_avatar(self, path: str):
        """
        A higher level interface to editing the bot's avatar.
        """
        return await self._bot.edit_avatar(path)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for custom emojis in guilds.

.. currentmodule:: curious.dataclasses.emoji
"""

import typing

from curious.dataclasses import guild as dt_guild, role as dt_role
from curious.dataclasses.bases import Dataclass


class Emoji(Dataclass):
    """
    Represents a custom emoji uploaded to a guild.
    """
    __slots__ = ("id", "name", "role_ids", "require_colons", "managed", "guild_id", "animated")

    def __init__(self, **kwargs):
        super().__init__(int(kwargs.get("id")), kwargs.get("client"))

        #: The name of this emoji.
        self.name = kwargs.get("name", None)  # type: str

        #: A list of role IDs that this emoji can be used by.
        self.role_ids = kwargs.get("roles", [])  # type: typing.List[int]

        #: If this emoji requires colons to use.
        self.require_colons = kwargs.get("require_colons", False)  # type: bool

        #: If this emoji is managed or not.
        self.managed = kwargs.get("managed", False)  # type: bool

        #: The ID of the guild this emoji is associated with.
        self.guild_id = None  # type: int

        #: If this emoji is animated or not.
        self.animated = kwargs.get("animated", False)  # type: bool

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return False

        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return "<{}:{}:{}>".format('a' if self.animated else '', self.name, self.id)

    def __repr__(self) -> str:
        return "<Emoji guild={!r} id={!r} name={!r}>".format(self.guild, self.id, self.name)

    async def edit(self, *,
                   name: str = None,
                   roles: 'typing.List[dt_role.Role]' = None) -> 'Emoji':
        """
        Edits this emoji.

        :param name: The new name of the emoji.
        :param roles: The new list of roles that can use this emoji.
        :return: This emoji.
        """
        if roles is not None:
            roles = [r.id for r in roles]

        await self._bot.http.edit_guild_emoji(guild_id=self.guild_id, emoji_id=self.id,
                                              name=name, roles=roles)
        return self

    async def delete(self) -> None:
        """
        Deletes this emoji.
        """
        await self._bot.http.delete_guild_emoji(self.guild_id, emoji_id=self.id)

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` this emoji object is associated with.
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def roles(self) -> 'typing.List[dt_role.Role]':
        """
        :return: A list of :class:`.Role` this emoji can be used by.
        """
        if len(self.role_ids) <= 0:
            return [self.guild.default_role]

        return [self.guild.roles[r_id] for r_id in self.role_ids]

    @property
    def url(self) -> str:
        """
        :return: The URL to this emoji.
        """
        cdn_url = f"https://cdn.discordapp.com/emojis/{self.id}"
        if not self.animated:
            return f"{cdn_url}.png"

        return f"{cdn_url}.gif"



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
A curio websocket wrapper.
"""

import threading
from typing import AsyncIterator

import curio
from curio.thread import AWAIT, async_thread
from lomond import WebSocket
from lomond.events import Event
from lomond.persist import persist

from curious import USER_AGENT
from curious.core._ws_wrapper import BasicWebsocketWrapper


class CurioWebsocketWrapper(BasicWebsocketWrapper):
    """
    Wraps a lomond websocket in a thread.
    """

    def __init__(self, url: str) -> None:
        super().__init__(url)

        #: The gateway task running in an async thread.
        self._task = None  # type: curio.Task

        self._queue = curio.UniversalQueue()
        self._cancelled = threading.Event()
        self._ws = None  # type: WebSocket

    @async_thread
    def websocket_handler(self) -> None:
        """
        The actual websocket handler.

        This wraps the Lomond reconnecting websocket, putting data on the queue.
        """
        ws = WebSocket(self.url, agent=USER_AGENT)
        self._ws = ws
        # generate poll events every 0.5 seconds to see if we can cancel
        websocket = persist(ws, ping_rate=0, poll=1, exit_event=self._cancelled)
        for event in websocket:
            self._queue.put(event)

        # signal the end of the queue
        self._queue.put(self._done)

    async def __aiter__(self) -> AsyncIterator[Event]:
        async for event in self._queue:
            if event is not self._done:
                yield event
            else:
                return

    @async_thread
    def send_text(self, message: str):
        """
        Sends text to the websocket.
        """
        self._ws.send_text(message)

    @async_thread
    def close(self, code: int = 1000, reason: str = "Client disconnect", reconnect: bool = False):
        """
        Cancels and closes this websocket.
        """

        # if reconnecting, don't close this as this will kill the websocket prematurely
        if not reconnect:
            self._cancelled.set()
            AWAIT(self._task.cancel(blocking=False))  # don't block because it closes by itself

        self._ws.close(code=code, reason=reason)

    @classmethod
    async def open(cls, url: str) -> 'BasicWebsocketWrapper':
        """
        Opens a websocket to the specified URL.
        """
        obb = cls(url)
        task = await curio.spawn(obb.websocket_handler())
        obb._task = task
        return obb



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Role objects.

.. currentmodule:: curious.dataclasses.role
"""

import functools

from curious.dataclasses import guild as dt_guild, member as dt_member, \
    permissions as dt_permissions
from curious.dataclasses.bases import Dataclass
from curious.exc import PermissionsError


class _MentionableRole(object):
    """
    A wrapper class that makes a role mentionable for a short time period.
    
    .. code-block:: python3
    
        async with role.allow_mentions():
            await ctx.channel.messages.send(role.mention)
            
    """
    def __init__(self, r: 'Role'):
        self.role = r

    def allow_mentions(self):
        return self.role.edit(mentionable=True)

    def disallow_mentions(self):
        return self.role.edit(mentionable=False)

    def __aenter__(self):
        return self.allow_mentions()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disallow_mentions()
        return False


@functools.total_ordering
class Role(Dataclass):
    """
    Represents a role on a server.
    """
    __slots__ = "name", "colour", "hoisted", "mentionable", "permissions", "managed", "position", \
                "guild_id"

    def __init__(self, client, **kwargs) -> None:
        super().__init__(kwargs.get("id"), client)

        #: The name of this role.
        self.name = kwargs.get("name", None)

        #: The colour of this role.
        self.colour = kwargs.get("color", 0)

        #: Is this role hoisted?
        self.hoisted = kwargs.get("hoist", False)

        #: Is this role mentionable?
        self.mentionable = kwargs.get("mentionable", False)

        #: The permissions of this role.
        self.permissions = dt_permissions.Permissions(kwargs.get("permissions", 0))

        #: Is this role managed?
        self.managed = kwargs.get("managed", False)

        #: The position of this role.
        self.position = kwargs.get("position", 0)

        #: The ID of the guild associated with this Role.
        self.guild_id = int(kwargs.get("guild_id", 0))  # type: dt_guild.Guild

    def __lt__(self, other: 'Role') -> bool:
        if not isinstance(other, Role):
            return NotImplemented

        if other.guild != self.guild:
            raise ValueError("Cannot compare roles between guilds")

        return self.position < other.position \
            if self.position != other.position \
            else self.id < other.id

    def _copy(self) -> 'Role':
        obb = object.__new__(self.__class__)

        obb.name = self.name
        obb.colour = self.colour
        obb.hoisted = self.hoisted
        obb.permissions = self.permissions
        obb.managed = self.managed
        obb.position = self.position
        obb.guild_id = self.guild_id

        return obb

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` associated with this role. 
        """
        return self._bot.guilds[self.guild_id]

    @property
    def is_default_role(self) -> bool:
        """
        :return: If this role is the default role of the guild.
        """
        return self.guild.id == self.id

    def allow_mentions(self) -> _MentionableRole:
        """
        Temporarily allows this role to be mentioned during.
         
        .. code-block:: python3
    
            async with role.allow_mentions():
                await ctx.channel.messages.send(role.mention)
        """
        return _MentionableRole(self)

    @property
    def mention(self) -> str:
        """
        Gets the string that can be used to mention this role. 
        
        .. warning::
        
            If :attr:`.Role.mentionable` is ``False``, this will not actually mention the role.
        
        """
        return f"<@&{self.id}>"

    async def assign_to(self, member: 'dt_member.Member') -> 'Role':
        """
        Assigns this role to a member.
        
        .. seealso::
        
            :meth:`.MemberRoleContainer.add`

        :param member: The :class:`.Member` to assign to.
        """
        await member.roles.add(self)
        return self

    async def remove_from(self, member: 'dt_member.Member'):
        """
        Removes this role from a member.
        
        .. seealso::
        
            :meth:`.MemberRoleContainer.remove`
        
        :param member: The :class:`.Member` to assign to.
        """
        await member.roles.remove(self)
        return self

    async def delete(self) -> 'Role':
        """
        Deletes this role.
        """
        if not self.guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        await self._bot.http.delete_role(self.guild.id, self.id)
        return self

    async def edit(self, *,
                   name: str = None, permissions: 'dt_permissions.Permissions' = None,
                   colour: int = None, position: int = None,
                   hoist: bool = None, mentionable: bool = None) -> 'Role':
        """
        Edits this role.

        :param name: The name of the role.
        :param permissions: The permissions that the role has.
        :param colour: The colour of the role.
        :param position: The position in the sorting list that the role has.
        :param hoist: Is this role hoisted (shows separately in the role list)?
        :param mentionable: Is this mentionable by everyone?
        """
        if not self.guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        if permissions is not None:
            if isinstance(permissions, dt_permissions.Permissions):
                permissions = permissions.bitfield

        async with self._bot.events.wait_for_manager("role_update", lambda b, a: a.id == self.id):
            await self._bot.http.edit_role(self.guild_id, self.id,
                                           name=name, permissions=permissions, colour=colour,
                                           hoist=hoist, position=position, mentionable=mentionable)
        return self



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Commands helpers.

.. currentmodule:: curious.commands

.. autosummary::
    :toctree: commands

    manager
    context
    decorators
    plugin
    utils
    ratelimit
    help

    exc
    converters
"""
from curious.commands.context import Context
from curious.commands.decorators import command, condition
from curious.commands.manager import CommandsManager
from curious.commands.plugin import Plugin



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
IPC helpers, for Rich Presence. This is NOT bot IPC.

.. currentmodule:: curious.ipc

.. autosummary::
    :toctree: ipc

    client
    packet
"""


"""
An example bot that uses events.
"""

# Events are the main way of listening to things that happen to the bot.
# They are registered with the @bot.event("name") decorator, or just the @event() decorator
# inside of plugins.

# Let's create a simple plugin that logs all messages, and another event that announces bans.

# First, the required imports
from curious.commands import CommandsManager
from curious.commands.plugin import Plugin
from curious.core.client import Client
from curious.core.event import EventContext, event
from curious.dataclasses.guild import Guild
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message


class BasicPlugin(Plugin):
    # We use the decorator to designate what event we wish to listen to.
    @event("message_create")
    # Events take at least one param - the EventContext. This contains our shard ID, as well as the
    # bot instance.
    async def log_message(self, ctx: EventContext, message: Message):
        # `log_message` takes a Message as its second argument, because it's a `message_create`
        # event.
        # Let's log the message content and author:
        print("Message recieved: `{}` from `{}`".format(message.content, message.author.user.name))
        # Let's also log the guild, if there is a guild.
        if message.guild is not None:
            print("Guild: {}".format(message.guild.name))
        # Finally, log the shard ID.
        print("Shard: {}".format(ctx.shard_id))


# Create the commands bot instance.
bot = Client()
manager = CommandsManager(client=bot, command_prefix="!")


# Add our plugin.
@bot.event("ready")
async def ready(ctx: EventContext):
    await manager.load_plugin(BasicPlugin)


# Now, we add the ban announcement event.
@bot.event("member_ban")
async def announce_ban(ctx: EventContext, guild: Guild, member: Member):
    # Send the ban message to the system channel.
    await guild.system_channel.messages.send("{} got bent".format(member.user.name))


# Now, all that is left is to run the bot.
bot.run('MjYwOTUwODE2NTM2NTI2ODQ5.Cz2mGQ.SKl78a6NT6SBpwYQrIDnR1olPqo')



import sys
from pathlib import Path

from setuptools import setup

install_requires = [
        "lomond>=0.1.13,<0.2",
        "pylru==1.0.9",
        "oauthlib>=2.0.2,<2.1.0",
        "pytz>=2017.3",
        "asks>=1.3.0,<1.4.0",
        "multidict>=4.1.0,<4.2.0",
        "multio>=0.2.1,<0.3.0",
        "async_generator~=1.9",  # asynccontextmanager for 3.6
        "typing_inspect>=0.2.0"
]

py36_requires = [
    "dataclasses>=0.3",  # PEP 557
    # pending 567 backport
]

if sys.version_info[0:2] <= (3, 6):
    install_requires += py36_requires


setup(
    name='discord-curious',
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "dirty-tag"
    },
    packages=['curious', 'curious.core', 'curious.core._ws_wrapper',
              'curious.commands', 'curious.dataclasses',
              'curious.ext.paginator', 'curious.ipc'],
    url='https://github.com/SunDwarf/curious',
    license='LGPLv3',
    author='Laura Dickinson',
    author_email='l@veriny.tf',
    description='An async library for the Discord API',
    long_description=Path(__file__).with_name("README.rst").read_text(encoding="utf-8"),
    python_requires=">=3.6.2",
    setup_requires=[
        "setuptools_scm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "Framework :: Trio",
        "Development Status :: 4 - Beta"
    ],
    install_requires=install_requires,
    extras_require={
        "voice": ["opuslib==1.1.0",
                  "PyNaCL==1.0.1"],
        "docs": [
            "sphinx_py3doc_enhanced_theme",
            "sphinx",
            "sphinxcontrib-asyncio",
            "sphinx-autodoc-typehints",
        ]
    },
)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Class for the commands context.

.. currentmodule:: curious.commands.context
"""
import inspect
import types
from typing import Any, Callable, List, Tuple, Type, Union

import typing_inspect

from curious.commands.converters import convert_channel, convert_float, convert_int, convert_list, \
    convert_member, convert_role
from curious.commands.exc import CommandInvokeError, CommandsError, ConditionsFailedError
from curious.commands.utils import _convert
from curious.core.event import EventContext
from curious.dataclasses.channel import Channel
from curious.dataclasses.guild import Guild
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.role import Role
from curious.dataclasses.user import User


class Context(object):
    """
    A class that represents the context for a command.
    """
    _converters = {
        Channel: convert_channel,
        Member: convert_member,
        Role: convert_role,
        # Guild: _convert_guild,
        List: convert_list,
        str: lambda ann, ctx, arg: arg,
        int: convert_int,
        float: convert_float,
    }

    def __init__(self, message: Message, event_context: EventContext):
        """
        :param message: The :class:`.Message` this command was invoked with.
        :param event_context: The EventContext for this context.
        """
        #: The message for this context.
        self.message = message  # type: Message

        #: The extracted command name for this context.
        self.command_name = None  # type: str

        #: The tokens for this context.
        self.tokens = []  # type: List[str]

        #: The formatted command for this context.
        self.formatted_command = None  # type: str

        #: The plugin for this context.
        self.plugin = None

        #: The manager for this context.
        self.manager = None

        #: The event context for this context.
        self.event_context = event_context  # type: EventContext

        #: The :class:`.Client` for this context.
        self.bot = event_context.bot

    @classmethod
    def add_converter(cls, type_: Type[Any], converter: 'Callable[[Context, str], Any]'):
        """
        Adds a converter to the mapping of converters.

        :param type_: The type to convert to.
        :param converter: The converter callable.
        """
        cls._converters[type_] = converter

    @property
    def guild(self) -> Guild:
        """
        :return: The :class:`.Guild` for this context, or None.
        """
        return self.message.guild

    @property
    def channel(self) -> Channel:
        """
        :return: The :class:`.Channel` for this context.
        """
        return self.message.channel

    @property
    def author(self) -> Union[Member, User]:
        """
        :return: The :class:`.Member` or :class:`.User` for this context.
        """
        return self.message.author

    def match_command(self, func) -> bool:
        """
        Attempts to match a command with this context.
        """
        # don't match subcommands
        if func.cmd_subcommand:
            return False

        # match on name
        if self.command_name == func.cmd_name:
            return True

        # match on alias
        if self.command_name in func.cmd_aliases:
            return True

        # no match
        return False

    def _lookup_converter(self, annotation: Type[Any]) -> 'Callable[[Any, Context, str], Any]':
        """
        Looks up a converter for the specified annotation.
        """
        origin = typing_inspect.get_origin(annotation)
        if origin is not None:
            annotation = origin

        if annotation in self._converters:
            return self._converters[annotation]

        if annotation is inspect.Parameter.empty:
            return lambda ann, ctx, i: i

        # str etc
        if callable(annotation):
            return annotation

        return lambda ann, ctx, i: i

    async def _get_converted_args(self, func) -> Tuple[tuple, dict]:
        """
        Gets the converted args and kwargs for this command, based on the tokens.
        """
        return await _convert(self, self.tokens, inspect.signature(func))

    def _make_reraise_ctx(self, new_name: str) -> EventContext:
        """
        Makes a new :class:`.EventContext` for re-dispatching.
        """
        return EventContext(self.bot, self.event_context.shard_id, new_name)

    async def _safety_wrapper(self, coro) -> None:
        """
        Runs a command in a safety wrapper.
        """
        evt_ctx = self._make_reraise_ctx("command_error")
        try:
            await coro
        except CommandsError as e:
            await self.manager.client.events.fire_event("command_error", self, e,
                                                        ctx=evt_ctx)
        except Exception as e:
            try:
                raise CommandInvokeError(self) from e
            except CommandInvokeError as e2:
                await self.manager.client.events.fire_event("command_error", self, e2,
                                                            ctx=evt_ctx)

    async def can_run(self, cmd) -> Tuple[bool, list]:
        """
        Checks if a command can be ran.

        :return: If it can be ran, and a list of conditions that failed.
        """
        conditions = getattr(cmd, "cmd_conditions", [])
        failed = []
        for condition in conditions:
            try:
                success = condition(self)
                if inspect.isawaitable(success):
                    success = await success
            except CommandsError:
                raise
            except Exception:
                failed.append(condition)
            else:
                if not success:
                    failed.append(success)

        if failed:
            return False, failed

        return True, []

    async def invoke(self, command) -> Any:
        """
        Invokes a command.
        This will convert arguments, pass them in to the command, and run the command.

        :param command: The command function to run.
        """
        # try and do a group lookup
        # how this works:
        # 1) it checks for the current command's subcommands
        # 1a) if empty, it assumes the current matched command is the best we can do
        #     and exits loop
        # 1b) if not empty, it checks all the subcommands for the current command
        # 2) when checking, it checks if the name of the subcommand matches
        # 3a) if it does, set current_command and matched_command, go back to 1
        # 3b) if it doesn't, exit loop so that current_command is the parent command
        matched_command = command
        current_command = command
        # used for subcommands only
        self_ = None
        if hasattr(current_command, "__self__"):
            self_ = current_command.__self__

        self.plugin = self_

        while True:
            if not current_command.cmd_subcommands:
                break

            if not self.tokens:
                break

            token = self.tokens[0]
            for command in current_command.cmd_subcommands:
                if command.cmd_name == token or token in command.cmd_aliases:
                    matched_command = command
                    current_command = command
                    # update tokens so that they're consumed
                    self.tokens = self.tokens[1:]
                    break
            else:
                # we didnt match any subcommand
                # so escape the loop now
                break

        # bind method, if appropriate
        if not hasattr(matched_command, "__self__") and self_ is not None:
            matched_command = types.MethodType(matched_command, self_)

        # check if we can actually run it
        can_run, conditions_failed = await self.can_run(matched_command)
        if not can_run:
            raise ConditionsFailedError(self, conditions_failed)

        # check if we're ratelimited
        await self.manager.ratelimiter.ensure_ratelimits(self, matched_command)

        # convert all the arguments into the command
        converted_args, converted_kwargs = await self._get_converted_args(matched_command)

        # finally, spawn the new command task
        try:
            return await matched_command(self, *converted_args, **converted_kwargs)
        except CommandsError:
            raise
        except Exception as e:
            raise CommandInvokeError(self) from e

    async def try_invoke(self) -> Any:
        """
        Attempts to invoke the command, using the specified manager.

        This will scan all the commands, then invoke as appropriate.
        """
        # temp variable used to invoke if applicable
        to_invoke = None

        for command in self.manager.commands.values():
            if self.match_command(command):
                to_invoke = command
                break

        for plugin in self.manager.plugins.values():
            commands = plugin._get_commands()
            for command in commands:
                if self.match_command(command):
                    to_invoke = command
                    break

        if to_invoke is not None:
            ev_ctx = self._make_reraise_ctx("command_error")
            try:
                return await self.invoke(to_invoke)
            except CommandsError as e:
                await self.manager.client.events.fire_event("command_error", self, e, ctx=ev_ctx)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Websocket gateway code.

.. currentmodule:: curious.core.gateway
"""
import collections
import enum
import logging
import sys
import threading
import time
import zlib
from contextlib import asynccontextmanager
from typing import Union, Coroutine, Any, AsyncGenerator

import curio
import multio
from asyncwebsockets import *
from curio.thread import AWAIT, async_thread

from curious.util import safe_generator

try:
    # Prefer ETF data.
    import earl
    _fmt = "etf"


    def _loader(data: bytes):
        return earl.unpack(data, encoding="utf-8", encode_binary_ext=True)


    def _dumper(data: dict):
        return earl.pack(data)

except ImportError:
    try:
        import ujson as json
    except ImportError:
        import json
    finally:
        _fmt = "json"


    def _loader(data: str):
        return json.loads(data)


    def _dumper(data: dict):
        return json.dumps(data)

from curious.dataclasses.presence import Game, Status


# Signalling exceptions.
class ChunkGuilds(Exception):
    """
    Signals that we need to begin downloading all guild member chunks.
    """


class GatewayError(OSError):
    """
    Raised when there's a gateway error.
    """


class GatewayOp(enum.IntEnum):
    """
    Represents the opcode mapping for the gateway.
    """
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE = 3
    VOICE_STATE = 4
    VOICE_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_MEMBERS = 8
    INVALIDATE_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11
    GUILD_SYNC = 12





@async_thread(daemon=True)
def _heartbeat_loop(gw: 'Gateway', heartbeat_interval: float):
    """
    Heartbeat looper that loops and sends heartbeats to the gateway.

    :param gw: The gateway to handle.
    """
    # async threads!
    gw.logger.debug("Sending initial heartbeat.")
    AWAIT(gw.send_heartbeat())
    while True:
        # this is similar to the normal threaded event waiter
        # it will time out after heartbeat_interval seconds.
        try:
            AWAIT(curio.timeout_after(heartbeat_interval, gw._stop_heartbeating.wait()))
        except curio.TaskTimeout:
            pass
        else:
            break

        try:
            AWAIT(gw.send_heartbeat())
        except ReconnectWebsocket:
            break


class Gateway(object):
    """
    Interacts with the Discord Gateway.
    """
    #: The current gateway version to connect to Discord via.
    GATEWAY_VERSION = 6

    def __init__(self, token: str, connection_state, *,
                 large_threshold: int = 250):
        """
        :param token: The bot token to connect with.
        """
        #: The token used to identify with.
        self.token = token

        #: The current websocket connection.
        self.websocket = None  # type: ClientWebsocket

        #: The current sequence number.
        self.sequence_num = 0

        #: The connection state used for this connection.
        self.state = connection_state

        #: The shard ID this gateway is representing.
        self.shard_id = 0
        #: The number of shards the client is connected to.
        self.shard_count = 1

        #: The current session ID for this gateway.
        self.session_id = None

        #: The current heartbeat statistic counter for this gateway.
        self.hb_stats = HeartbeatStats()

        #: The current game for this gateway.
        #: Only set when sending a change status packet.
        self.game = None
        #: The current status for this gateway.
        self.status = None

        #: The "large threshold" for this Gateway.
        #: For guilds with members above this number, offline members will not be sent.
        self.large_threshold = min(250, large_threshold)  # bound to 250

        #: The data format for this gateway.
        self.format = _fmt

        self._tg = None
        self._prev_seq = 0
        self._dispatches_handled = collections.Counter()
        self._enqueued_guilds = []
        self._stop_heartbeating = multio.Event()
        self._logger = None
        self._cached_gateway_url = None  # type: str
        self._open = False
        self._close_code = None
        self._close_reason = None

    @property
    def logger(self):
        if self._logger:
            return self._logger

        self._logger = logging.getLogger("curious.gateway:shard-{}".format(self.shard_id))
        return self._logger

    async def connect(self, url: str):
        """
        Connects the client to the websocket.

        This does not do ANY handshake steps.
        :param url: The URL to connect to.
        """
        self.logger.info("Opening connection to {}".format(url))
        try:
            self.websocket = await curio.timeout_after(5, connect_websocket(url))
        except multio.asynclib.TaskTimeout as e:
            raise ReconnectWebsocket from e
        self.logger.info("Connected to gateway!")
        self._open = True
        return self.websocket

    async def _start_heartbeating(self, heartbeat_interval: float) -> threading.Thread:
        """
        Starts the heartbeat thread.
        :param heartbeat_interval: The number of seconds between each heartbeat.
        """
        if not self._stop_heartbeating.is_set():
            # cancel some poor thread elsewhere
            await self._stop_heartbeating.set()
            del self._stop_heartbeating

        self._stop_heartbeating = multio.Event()

        # dont reference the task - it'll die by itself
        task = await curio.spawn(_heartbeat_loop(self, heartbeat_interval), daemon=True)

        return task

    # Utility methods.
    async def close(self, code: int = 1000, reason: str = "Client disconnected"):
        """
        Closes the connection.

        :param code: The close code.
        :param reason: The close reason.
        """
        if not self.websocket.closed:
            await self.websocket.close(code=code, reason=reason)

        self._open = False
        self._close_code = code
        self._close_reason = reason
        await self._stop_heartbeating.set()

    async def _send(self, data: Union[bytes, str]) -> None:
        """
        Sends some data.

        This will actually send the data down the websocket, unlike `send` which only pretends to.
        """
        # Check if heartbeats have drifted.
        if self.hb_stats.heartbeat_acks + 2 < self.hb_stats.heartbeats:
            self.logger.error("Heartbeats have drifted, closing connection!")
            await self.close(code=1000, reason="Failed to receive heartbeat ACKs in time")
            raise ReconnectWebsocket

        await self.websocket.send_message(data)

    def _send_dict(self, payload: dict) -> Coroutine[None, None, None]:
        """
        Sends a dict to be packed down the gateway.

        This is a private method - use :meth:`send` instead.

        :param payload: The payload to send.
        """
        data = _dumper(payload)
        return self._send(data)

    async def send(self, data: Any) -> None:
        """
        Sends a variable type of data down the gateway.
        """
        if isinstance(data, dict):
            await self._send_dict(data)
        else:
            await self._send(data)

    # Sending events.
    async def send_identify(self, os: str = sys.platform, browser: str = "curious",
                            device: str = "curious", compress: bool = True) -> None:
        """
        Sends an IDENTIFY packet.

        :param os: The ``$os`` property to send when identifying. Defaults to sys.platform.
        :param browser: The ``$browser`` property to send when identifying. Defaults to ``curious``.
        :param device: The ``$device`` property to send when identifying. Defaults to ``curious``.
        :param compress: Should the connection be compressed? Defaults to True.
        """
        payload = {
            "op": GatewayOp.IDENTIFY,
            "d": {
                "token": self.token,
                "properties": {
                    "$os": os,
                    "$browser": browser,
                    "$device": device,
                    "$referrer": "",
                    "$referring_domain": ""
                },
                "compress": compress,
                "large_threshold": self.large_threshold,
                "v": self.GATEWAY_VERSION,
                "shard": [self.shard_id, self.shard_count]
            }
        }

        await self._send_dict(payload)

    async def send_resume(self) -> None:
        """
        Sends a RESUME packet.
        """
        payload = {
            "op": GatewayOp.RESUME,
            "d": {
                "token": self.token,
                "session_id": self.session_id,
                "seq": self.sequence_num
            }
        }

        await self._send_dict(payload)

    async def send_status(self, game: Game, status: Status, *,
                          afk: bool=False) -> None:
        """
        Sends a PRESENCE_UPDATE packet.

        :param game: The game object to send.
        :param status: The status object to send.
        :param afk: Is this gateway AFK?
            This should be True for self bots.
        """
        payload = {
            "op": GatewayOp.PRESENCE,
            "d": {
                "game": game.to_dict() if game else None,
                "status": status.value,
                "afk": afk,
                "since": None if not afk else int(time.time() * 1000)
            }
        }

        self.game = game
        self.status = status

        # Update our game() object on all guilds on this shard.
        for guild in self.state.guilds_for_shard(self.shard_id):
            try:
                guild.me.presence.game = game
                guild.me.presence.status = status
            except AttributeError:
                # sent before our member object exists - i.e just after READY happens
                # we can ignore this
                pass

        await self._send_dict(payload)

    async def send_voice_state_update(self, guild_id: int, channel_id: int) -> None:
        """
        Sends a voice state update packet.

        :param guild_id: The guild ID to update in.
        :param channel_id: The channel ID to update in.
        """

        payload = {
            "op": GatewayOp.VOICE_STATE,
            "d": {
                "guild_id": str(guild_id),
                "channel_id": str(channel_id),
                "self_mute": False,
                "self_deaf": False,
            }
        }

        await self._send_dict(payload)

    async def send_heartbeat(self) -> int:
        """
        Sends a single heartbeat.
        """
        hb = self._get_heartbeat()
        self.logger.debug("Heartbeating with sequence {}".format(hb["d"]))

        # increment the stats
        self.hb_stats.heartbeats += 1
        self.hb_stats.last_heartbeat = time.monotonic()

        await self._send_dict(hb)
        return self.hb_stats.heartbeats

    async def send_guild_sync(self, guilds) -> None:
        """
        Sends a guild sync packet.

        :param guilds: A list of guild IDs to request syncing for.
        """
        payload = {
            "op": GatewayOp.GUILD_SYNC,
            "d": [str(g.id) for g in guilds]
        }

        await self._send_dict(payload)

    async def request_chunks(self, guilds) -> None:
        """
        Requests member chunks from a guild.

        :param guilds: A list of guild IDs to request chunks for.
        """
        payload = {
            "op": GatewayOp.REQUEST_MEMBERS,
            "d": {
                "guild_id": [str(guild.id) for guild in guilds],
                "query": "",
                "limit": 0  # Request ALL!
            }
        }

        return await self._send_dict(payload)

    @classmethod
    def from_token(cls, token: str, state, gateway_url: str,
                         *, shard_id: int = 0, shard_count: int = 1,
                         **kwargs):
        """
        Creates a new gateway connection from a token.

        :param token: The token to pass in.
            This must be a bot user token - user tokens are not supported.
        :param state: The connection state that is attached to this client.
        :param gateway_url: The gateway URL to connect.
        :param shard_id: The shard ID of this bot.
        :param shard_count: The number of shards to start the bot with.

        :return: A new :class:`Gateway` that is connected to the API.
        """
        obb = cls(token, state, **kwargs)
        obb.shard_id = shard_id
        obb.shard_count = shard_count

        gateway_url += "/?v={}&encoding={}".format(cls.GATEWAY_VERSION, _fmt)
        obb._cached_gateway_url = gateway_url

        return obb.run()

    async def reconnect(self, *, resume: bool = False) -> 'Gateway':
        """
        Reconnects the bot to the gateway.

        :param resume: Should a RESUME be attempted?
        """
        self.logger.info("Reconnecting to the gateway")

        # reset our heartbeat count
        self.hb_stats.heartbeats = 0
        self.hb_stats.heartbeat_acks = 0

        if not self.websocket.closed:
            await self.websocket.close(code=1001, reason="Forcing a reconnect")

        self._open = False

        await self.connect(self._cached_gateway_url)

        if resume:
            # Send the RESUME packet, instead of the IDENTIFY packet.
            self.logger.info("Sending RESUME...")
            await self.send_resume()
        else:
            self.logger.info("Sending IDENTIFY...")
            self.sequence_num = 0
            await self.send_identify()

        return self

    async def _get_chunks(self) -> None:
        """
        Called to start chunking all guild members.
        """
        for guild in self._enqueued_guilds:
            guild.start_chunking()
            self.logger.info("Requesting {} member chunk(s) from guild {}."
                             .format(guild._chunks_left, guild.name))

        await self.request_chunks(self._enqueued_guilds)
        self._enqueued_guilds.clear()

    def _get_heartbeat(self) -> dict:
        return {
            "op": GatewayOp.HEARTBEAT,
            "d": self.sequence_num
        }

    @asynccontextmanager
    @safe_generator
    async def run(self):
        """
        Runs the websocket, identifying as appropriate.

        Correct usage of this function::

            async with websocket.run() as gen:
                async for event in gen:
                    ...

        This will automatically handle connecting and reconnecting.
        """
        async with multio.asynclib.task_manager() as tg:
            try:
                self._tg = tg
                gen = self._run_generator()
                async with multio.asynclib.finalize_agen(gen) as finalized:
                    yield finalized

            finally:
                await multio.asynclib.cancel_task_group(tg)
                self._tg = None

    async def _run_generator(self):
        """
        Represents the run generator; this is the actual runner implementation for the gateway
        generator.
        """
        ws = await connect_websocket(self._cached_gateway_url, reconnecting=True)
        self.websocket = ws
        async for event in ws:
            # we can't start heartbeating when we open unfortunately
            if isinstance(event, WebsocketClosed):
                await self._close_heartbeat_task()

                # check to see if we can handle it
                if event.code in (4004, 4011):
                    # 4004 -> invalid token, 4011 -> too many guilds
                    await ws.close(code=1006, reason="Unhandled client error")
                    raise GatewayError(event.code, event.reason)

            elif isinstance(event, (WebsocketTextMessage, WebsocketBytesMessage)):
                async with multio.asynclib.finalize_agen(self._handle_data_message(event)) as gen:
                    async for evt in gen:
                        yield evt

    async def _handle_data_message(self,
                                   event: Union[WebsocketBytesMessage, WebsocketTextMessage]) \
        -> AsyncGenerator[tuple, None]:
        """
        Creates an async generator that yields new events to be dispatched.
        """
        if not self._open:
            raise WebsocketClosed(1006, reason="Connection lost")

        yield ("gateway_message_received", event)

        # decompress the data, if needed
        if isinstance(event, WebsocketBytesMessage) and _fmt == "json":
            data = zlib.decompress(event.data, 15, 10490000)
            data = data.decode("utf-8")
        else:
            data = event.data

        # skip empty payloads
        if not data:
            return

        # load the event data with our loader
        event_data = _loader(data)
        yield ("gateway_event_received", event_data)

        op = event_data.get('op')
        data = event_data.get('d')
        seq = event_data.get('s')

        if seq is not None:
            # next sequence
            self.sequence_num = seq

        # Handle internal operations, as well as dispatches.
        if op == GatewayOp.HELLO:
            # Start heartbeating, with the specified heartbeat duration.
            yield ("gateway_hello", data['_trace'])

            heartbeat_interval = data.get("heartbeat_interval", 45000) / 1000.0

            self.logger.debug("Heartbeating every {} seconds.".format(heartbeat_interval))
            await self._open_heartbeat_task(heartbeat_interval)
            self.logger.info("Connected to Discord servers {}".format(",".join(data["_trace"])))

            if self.session_id is None:
                self.logger.info("Sending an IDENTIFY...")
                await self.send_identify()
            else:
                self.logger.info("Sending a RESUME...")
                await self.send_resume()

        elif op == GatewayOp.HEARTBEAT_ACK:
            yield "gateway_heartbeat_ack",
            self.hb_stats.heartbeat_acks += 1
            self.hb_stats.last_ack = time.monotonic()

        elif op == GatewayOp.HEARTBEAT:
            # Send a heartbeat back.
            yield "gateway_heartbeat_received",
            await self.send_heartbeat()

        elif op == GatewayOp.INVALIDATE_SESSION:
            # the data sent is if we should resume
            # if it's non-existent, we assume it's False.
            should_resume = data or False

            yield ("gateway_invalidate_session", should_resume,)
            if should_resume is True:
                self.logger.debug("Sending RESUME again")
                await self.send_resume()
            else:
                self.logger.warning("Received INVALIDATE_SESSION with d False, re-identifying.")
                self.sequence_num = 0
                self.state._reset(self.shard_id)
                # clear session id so that if we reconnect we
                self.session_id = None
                await self.send_identify()

        elif op == GatewayOp.RECONNECT:
            yield ("gateway_reconnect_received",)
            await self.close(code=1000, reason="reconnect issued")

        elif op == GatewayOp.DISPATCH:
            # Handle the dispatch.
            event = event_data.get("t")
            self._dispatches_handled[event] += 1
            yield ("gateway_dispatch_received", event, data,)

        else:
            try:
                self.logger.warning("Unhandled opcode: {} ({})".format(op, GatewayOp(op)))
            except ValueError:
                self.logger.warning("Unknown opcode: {}".format(op))


# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

from curious.dataclasses.bases import Dataclass


class Attachment(Dataclass):
    """
    Represents an attachment.
    """
    def __init__(self, id: int, bot, **kwargs):
        super().__init__(id, cl=bot)

        #: The filename for this attachment.
        self.filename: str = kwargs.get("filename")

        #: The size of this attachment (in bytes).
        self.size: int = kwargs.get("size")

        #: The URL of this attachment.
        self.url: str = kwargs.get("url")

        #: The proxy_url of this attachment.
        self.proxy_url: str = kwargs.get("proxy_url")

        #: The height of this attachment, if an image.
        self.height: int = kwargs.get("height")

        #: The width of this attachment, if an image.
        self.width: int = kwargs.get("width")

    async def download(self) -> bytes:
        """
        Downloads the attachment into bytes.
        """
        bucket = ("attachment", self.id)
        data = await self._bot.http.request(bucket, method="GET", uri=self.proxy_url)
        return data



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Webhook objects.

.. currentmodule:: curious.dataclasses.webhook
"""

import typing

from curious.dataclasses import channel as dt_channel, embed as dt_embed, guild as dt_guild, \
    user as dt_user
from curious.dataclasses.bases import Dataclass
from curious.util import base64ify


class Webhook(Dataclass):
    """
    Represents a webhook on the guild.
    Messages in a guild can be sent by either a Member or a Webhook object - curious makes a key
    distinction between them. These classes are *mostly* compatible and don't require much
    effort to use them generically.
    
    .. code-block:: python3
    
        @event("message_create")
        async def handle_messages(ctx, message: Message):
            author = message.author  # can be Webhook or Member
    """

    __slots__ = "user", "guild_id", "channel_id", "token", "owner", \
                "default_name", "_default_avatar"

    def __init__(self, client, **kwargs) -> None:
        # Use the webhook ID is provided (i.e created from a message object).
        # If that doesn't exist, we use the ID of the data instead (it's probably right!).
        super().__init__(kwargs.get("webhook_id", kwargs.get("id")), cl=client)

        #: The user object associated with this webhook.
        self.user = None  # type: dt_user.User

        #: The ID of the Guild associated with this object.
        self.guild_id = None  # type: int

        #: The ID of the Channel associated with this object.
        self.channel_id = None  # type: int

        #: The token associated with this webhook.
        #: This is None if the webhook was received from a Message object.
        self.token = kwargs.get("token", None)  # type: str

        #: The owner of this webhook.
        self.owner = None  # type: dt_user.User

        #: The default name of this webhook.
        self.default_name = None  # type: str

        #: The default avatar of this webhook.
        self._default_avatar = None  # type: str

    def __repr__(self) -> str:
        return "<Webhook id={} name={} channel={} owner={}>".format(self.id, self.name,
                                                                    repr(self.channel),
                                                                    repr(self.owner))

    __str__ = __repr__

    @property
    def default_avatar_url(self) -> str:
        """
        :return: The default avatar URL for this webhook.
        """
        return "https://cdn.discordapp.com/avatars/{}/{}.png".format(self.id, self._default_avatar)

    @property
    def avatar_url(self) -> str:
        """
        :return: The computed avatar URL for this webhook.
        """
        if self.user.avatar_hash is None:
            return self.default_avatar_url
        return str(self.user.avatar_url)

    @property
    def name(self) -> str:
        """
        :return: The computed name for this webhook.
        """
        # this is kept so you can easily do `message.author.name` all the time.
        return self.user.name or self.default_name

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` this webhook is in.
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def channel(self) -> 'dt_channel.Channel':
        """
        :return: The :class:`.Channel` this webhook is in.
        """
        if self.guild is None:
            return None

        return self.guild.channels.get(self.channel_id)

    @classmethod
    async def create(cls, channel: 'dt_channel.Channel', *,
                     name: str, avatar: bytes) -> 'Webhook':
        """
        Creates a new webhook.

        :param channel: The :class:`.Channel` to create the webhook in.
        :param name: The name of the webhook to create.
        :param avatar: The bytes data for the webhook's default avatar.
        :return: A new :class:`.Webhook`.
        """
        return await channel.create_webhook(name=name, avatar=avatar)

    async def get_token(self) -> str:
        """
        Gets the token for this webhook, if no token was set earlier.

        :return: The token for the webhook.
        """
        if self.token:
            return self.token

        us = await self._bot.http.get_webhook(self.id)
        self.token = us.get("token")
        return self.token

    async def delete(self) -> None:
        """
        Deletes the webhook.

        You must either be the owner of this webhook, or the webhook must have a token associated
        to delete it.
        """
        if self.token is not None:
            return await self._bot.http.delete_webhook_with_token(self.id, self.token)
        else:
            return await self.guild.delete_webhook(self)

    async def edit(self, *,
                   name: str = None, avatar: bytes = None) -> 'Webhook':
        """
        Edits this webhook.

        :param name: The new name for this webhook.
        :param avatar: The bytes-encoded content of the new avatar.
        :return: The webhook object.
        """
        if avatar is not None:
            avatar = base64ify(avatar)

        if self.token is not None:
            # edit with token, don't pass to guild
            data = await self._bot.http.edit_webhook_with_token(self.id, name=name, avatar=avatar)
            self.default_name = data.get("name")
            self._default_avatar = data.get("avatar")

            # Update the user too
            self.user.username = data.get("name")
            self.user.avatar_hash = data.get("avatar")
        else:
            await self.channel.edit_webhook(self, name=name, avatar=avatar)

        return self

    async def execute(self, *,
                      content: str = None, username: str = None, avatar_url: str = None,
                      embeds: 'typing.List[dt_embed.Embed]'=None, wait: bool = False) \
            -> typing.Union[None, str]:
        """
        Executes the webhook.

        :param content: Any raw content to send.
        :param username: A username to override the default username of the webhook with.
        :param avatar_url: The URL for the avatar to override the default avatar with.
        :param embeds: A list of embeds to add to the message.
        :param wait: Should we wait for the message to arrive before returning?
        """
        if embeds:
            embeds = [embed.to_dict() for embed in embeds]

        if self.token is None:
            await self.get_token()

        data = await self._bot.http.execute_webhook(self.id, self.token,
                                                    content=content, embeds=embeds,
                                                    username=username, avatar_url=avatar_url,
                                                    wait=wait)

        if wait:
            return self._bot.state.make_message(data, cache=False)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Reaction objects.

.. currentmodule:: curious.dataclasses.reaction
"""

import typing

from curious.dataclasses import emoji as dt_emoji


class Reaction(object):
    """
    Represents a reaction.
    """
    def __init__(self, **kwargs) -> None:
        #: The :class:`.Message` this reaction is for.
        self.message = None

        #: The emoji that represents this reaction.
        self.emoji = None  # type: typing.Union[str, dt_emoji.Emoji]

        #: The number of times this message was reacted to.
        self.count = kwargs.get("count", 1)  # 1 is better than 0

        #: If this user reacted to the message.
        self.me = kwargs.get("me", False)

    def __repr__(self) -> str:
        return "<Reaction emoji={} count={}>".format(self.emoji, self.count)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Reaction):
            return NotImplemented

        if self.message.id != other.message.id:
            return NotImplemented

        return self.emoji == other.emoji

    def __hash__(self) -> int:
        # naiive
        return hash(self.message) + hash(self.emoji)



"""
Permissions pyi file.
"""
from typing import Any, Union

from curious.dataclasses.channel import Channel
from curious.dataclasses.member import Member
from curious.dataclasses.role import Role


class Permissions(object):
    __slots__ = 'bitfield',
    def __new__(cls, value: int, **kwargs): ...
    def __init__(self, value: int, **kwargs):
        self.bitfield: int = value

    def _get_bit(self, bit: int) -> bool: ...
    def _set_bit(self, bit: int, value: bool): ...
    def __repr__(self) -> str: ...

    @classmethod
    def all(cls) -> Permissions: ...
    @classmethod
    def none(cls) -> Permissions: ...

    def raise_for_permission(self, permission: Union[str, int]) -> None: ...

    @property
    def create_instant_invite(self) -> bool: ...
    @property
    def kick_members(self) -> bool: ...
    @property
    def ban_members(self) -> bool: ...
    @property
    def administrator(self) -> bool: ...
    @property
    def manage_channels(self) -> bool: ...
    @property
    def manage_server(self) -> bool: ...
    @property
    def add_reactions(self) -> bool: ...
    @property
    def view_audit_log(self) -> bool: ...
    @property
    def read_messages(self) -> bool: ...
    @property
    def send_messages(self) -> bool: ...
    @property
    def send_tts_messages(self) -> bool: ...
    @property
    def manage_messages(self) -> bool: ...
    @property
    def embed_links(self) -> bool: ...
    @property
    def attach_files(self) -> bool: ...
    @property
    def read_message_history(self) -> bool: ...
    @property
    def mention_everyone(self) -> bool: ...
    @property
    def use_external_emojis(self) -> bool: ...
    @property
    def voice_connect(self) -> bool: ...
    @property
    def voice_speak(self) -> bool: ...
    @property
    def voice_mute_members(self) -> bool: ...
    @property
    def voice_deafen_members(self) -> bool: ...
    @property
    def voice_move_members(self) -> bool: ...
    @property
    def voice_use_voice_activation(self) -> bool: ...
    @property
    def change_nickname(self) -> bool: ...
    @property
    def manage_nicknames(self) -> bool: ...
    @property
    def manage_roles(self) -> bool: ...
    @property
    def manage_webhooks(self) -> bool: ...
    @property
    def manage_emojis(self) -> bool: ...


class Overwrite(object):
     __slots__ = "target", "channel_id", "allow", "deny"

     def __init__(self, allow: Union[int, Permissions], deny: Union[int, Permissions],
                  obb: Union[Member, Role],
                  channel_id: int = None):
         self.allow = allow
         self.deny = deny
         self.target = obb
         self.channel_id = channel_id
     def __repr__(self) -> str: ...
     def __getattr__(self, item) -> bool: ...
     def __setattr__(self, item: str, value: Any): ...
     @property
     def channel(self) -> Channel: ...